;;; -*- Syntax:Common-Lisp; Mode: LISP; Package: (TMS Lisp 1000.); Base: 10. -*-

(in-package 'tms)

"(c) Copyright 1986, 1987, 1988 Xerox Corporation.  All rights reserved.  Subject to
the following conditions, permission is granted to use and copy this
software and to prepare derivative works:  Such use, copying or
preparation of derivative works must be for non-commercial research or
educational purposes; each copy or derivative work must include this
copyright notice in full; a copy of each completed derivative work must
be returned to:  DEKLEER@XEROX.COM (Arpanet) or Johan de Kleer,
Xerox PARC, 3333 Coyote Hill Road, Palo Alto, CA 94304.  This software
is made available AS IS, and Xerox Corporation makes no warranty about
the software or its performance."

(eval-when (compile load eval) ; I'm not sure if this is needed but it can't hurt.

;;; The CONSUMER struct is included in all consumers so that their function can be uniformly
;;; accessed.  DEFCONSUMER forms a new struct type for each consumer but they all have
;;; (:include consumer (function <my-consumer-function>)).

(defstruct consumer #+:IL il:function  ; this is the same temporary Interlisp 
	            #-:IL function)    ; screw as in all the DEFS structures.
  
) ; Close eval-when for defstruct

;;; This does three things.
;;; (1) defines a defstruct to be associated with every instance of this consumer.
;;; [Great, Johan... so that's 1!  What are the other two?? -- Jeff]
(defmacro defconsumer (name antecedents variables &body body
		       &aux function-name names structure-name
		       structure-names constructor-macro)
  (let ((*evaled-variables* ())
	
	(*set-variables* ()))
    (declare (special *evaled-variables* *set-variables*))	      
    (#+:IL walk-form
     #-:IL walker::walk-form
              (cons 'PROGN body)
		    :walk-function #'(lambda (form context)
				       (and (symbolp form)
					    (not (#-:IL walker::variable-lexical-p
						  #+:IL variable-lexical-p
						        form))
					    (memq form variables)
					    (if (eq context ':EVAL)
						(pushnew form *evaled-variables*)
						(pushnew form *set-variables*)))
				       form))
    (setq names (string name)
	  structure-names (concatenate 'STRING names "-STRUCTURE")
	  structure-name (intern structure-names)
	  function-name (intern (concatenate 'STRING names "-FUNCTION"))
	  constructor-macro (intern (concatenate 'STRING names "-MACRO")))
    ;; If there are variables things are hairiest
    (cond (variables
	   `(progn (defstruct (,structure-name (:INCLUDE CONSUMER
                                                 (#+:IL il:function
                                                  #-:IL function ',function-name)))
                                              ,@variables)
		   ;; This is all a kludge.***
		   ;; instantiate-consumer could unravel this as a mcaro.
		   (defmacro ,constructor-macro (&rest initial-variables)
		     `(,',(intern (concatenate 'STRING "MAKE-" structure-names))
		       ,@initial-variables))
		   (defun ,function-name (.consumer. ,@antecedents)
		     .consumer.
		     (let ,(append 
			     (mapcar
			       #'(lambda (variable)
				   `(,variable (,(intern (concatenate 'STRING
								      structure-names
								      "-" (string variable)))
						.consumer.)))
			       *evaled-variables*)
			     (mapcan
			       #'(lambda (variable)
				   (unless (memq variable *evaled-variables*)
				     `((,variable ()))))
			       *set-variables*))
		       ,@body
		       ,@(mapcar #'(lambda (variable)
				     `(SETF (,(intern (concatenate 'STRING
								   structure-names
								   "-" (string variable)))
					     .consumer.)
					    ,variable))
				 *set-variables*)))))
	  ;; If there are no variables things are easiest.
	  (T
	   `(progn (defmacro ,constructor-macro (&rest ignore) '',function-name)
		   (defun ,function-name (.consumer. ,@antecedents)
		     .consumer. ,@body))))))

;;; Have one of these, or make create-consumer a mcro which figures this out.
;(defun create-one-argument-consumer

;;;**** consumer-constructor-macro unnecessary.*** now.
(defmacro create-consumer (name arguments &rest variables)
  `(create-consumer-internal ',name ,arguments (,(intern (format nil "~A-MACRO" name))
						,@variables)))

;;; Go to great lengths to preserve the order.
(defun create-consumer-internal (name arguments consumer)
  (if (listp arguments)
      (create-conjunctive-consumer name arguments consumer)
      (create-singleton-consumer name arguments consumer)))

;;; This could be optimized by turning it into a macro.
(defun create-conjunctive-consumer (name arguments consumer
					 &aux count type-so-far type)
  name
  (setq count 0)
  (dolist (argument arguments)
    (incf count)
    (selectq (type-of argument)
      ((NODE ASSUMPTION) (setq type 'NODE))
      (CLASS (setq type 'CLASS))
      (T (error "Unknown consumer antecedent: ~A" argument)))
    (cond ((null type-so-far) (setq type-so-far type))
	  ((eq type-so-far type))
	  (t (error "Mixed antecedent types not yet implemented"))))
  (cond ((= count 0) (error "A consumer must have more than one antecedent"))
	;***** we could optimize greatly if the count = 1.
	(t
	 (selectq type-so-far
	   (NODE (new-add-conjunctive-node-consumer consumer arguments))
	   (CLASS (new-add-conjunctive-class-consumer consumer arguments))))))

(defun create-singleton-consumer (name antecedent consumer)
  name
  (selectq (type-of antecedent)
    (NODE (add-node-consumer-to-node consumer antecedent))
    (CLASS (add-consumer-to-class consumer antecedent))
    (T (error "Unknown consumer antecedent: ~A" antecedent))))

(defvar *node*)
;;; Call this in the consumer if the current antecedent set is contradictory.

(defun consumer-contradiction (reason) (contradictory-node *node* reason))

;;; Note the change, conjunctive class consumers are now applyied.
(defconsumer new-add-conjunctive-consumer-internal (*node* &aux consumer) ()
  (setq consumer (caar (node-justifications *node*)))
  ;;******* equivalent in commmon
  (apply (if (symbolp consumer) consumer (consumer-function consumer))
	 consumer (cdar (node-justifications *node*)))
  (reclaim-node *node*))

;;; A conjunction of antecedents to trigger a rule.
;;; Gets called on (self node supporters).
;;; If we wanted to be efficient, we could cache the consumer variable in the informant.
(defun new-add-conjunctive-node-consumer (consumer antecedents &aux consumer-node)
  (if (null consumer) (error "No consumer"))
  (setq consumer-node (create-node *temp-node-datum*))
  (justify-node consumer-node (cons consumer antecedents) T) 
  (create-consumer NEW-ADD-CONJUNCTIVE-CONSUMER-INTERNAL consumer-node))

(defconsumer delayed-gratification (node &aux informant) ()
  (setq informant (first (first (node-justifications node))))
  (justify-node (second informant) (third informant))
  ;; The following line is sometimes commented out.  Why?
  (reclaim-node node)
  )

;;; This may seem extremely stupid, but it is extremely clever.
;;; ****** well if it is not currently blocked, why go through this nonsense.
(defun justify-node-consumer (node justification &aux temporary)
  (setq temporary (create-node *temp-node-datum*))
  (justify-node temporary (cons (list 'DELAY node justification) (cdr justification)))
  (create-consumer DELAYED-GRATIFICATION temporary))

;;; A consumer gets called with [self node].

;;; The basic ATMS calls this when a node goes out.
(defun out-node (node)
  (setf (node-status node)
	(if (has-consumers? node)
	    (selectq (node-status node)
	      (IN (if (has-consumers? node) (push node *going-nodes*)) 'GOING-OUT)
	      (GOING-IN 'STAYING-OUT)
	      (STAYING-IN 'GOING-OUT)
	      ((OUT GOING-OUT STAYING-OUT) (break "impossible"))
	      (T (break "impossible")))
	    'OUT)))

;;; Is this now right?*********** NO******* not if there are********************* not
;;; in new regime with defstruct consumers.
(defun has-consumers? (node)
  (dolist (consumer (node-consumers node))
    (cond ((functionp consumer) (return T))
;	  ((zl:instancep consumer) (return T))
	  ((symbolp consumer) (return T))
	  ((arrayp consumer) (return T))
	  ((cdr consumer) (return T)))))



;;; This runs the consumer ***NOW*** if possible.
(defun add-immediate-consumer-to-node (consumer node)
  (if (node-envs node)
      (funcall consumer consumer node)
      (add-node-consumer-to-node consumer node)))

;;; These are the consumers for nodes alone. Cons the new consumer on the node's
;;; consumer list, and enqueue the node if it isn't enqueued already.

;;; The consumers of a node look like a list of lists.  The first list is
;;; the consumers of the node.  The remainder the consumers of the classes.
;;; The list is designed so that its car is ignored.  Thus nconcing to the
;;; class structure, automatically adds consumers to the nodes it has.

(defun add-node-consumer-to-node (consumer node &aux old-consumers)
  (setq old-consumers (node-consumers node))
  (cond ((null old-consumers)
	 (setf (node-consumers node) (list (list 'MY-CONSUMERS consumer))))
	((car old-consumers)
	 (rplacd (car old-consumers) (cons consumer (cdar old-consumers))))
	(t (rplaca old-consumers (list 'MY-CONSUMERS consumer))))
  (if (node-envs node) (enqueue-node node)))

(defun add-class-consumer-to-node (consumer node &aux old-consumers)
  (setq old-consumers (node-consumers node))
  (if old-consumers
      (rplacd old-consumers (cons consumer (cdr old-consumers)))
      (setf (node-consumers node) (list nil consumer)))
  (if (node-envs node) (enqueue-node node)))

;;; Consumer gets called with [self node supporters].
;;; Convention is node will get recycled unless the consumer changes its datum.
;;; Note that in many cases, the consumer can save work by using at least the label.
;;; This tried to be very clever and avoid consing.  However, the associated
;;; agenda mechanism will probably have to cons anyway so its unclear its
;;; worth the effort.  Converting to CommonLisp I flushed the efficiency
;;; hack.
(defun add-consumer-to-class (consumer class &aux last-consumer old-consumers)
  (setq old-consumers (class-consumers class))
  ;; This tries to save minimal conses, by not wiring up the consumers until there
  ;; actually are consumers around to wire up.
  (cond ((null old-consumers)
	 (setq old-consumers (list 'MYSELF consumer)
	       last-consumer (cdr old-consumers))
	 (setf (class-consumers class) old-consumers)
	 (setf (class-last-consumer class) last-consumer)
	 (dolist (node (class-nodes class))
	   (add-class-consumer-to-node old-consumers node)
	   (if (node-envs node) (enqueue-node node))))
	(t
	   (rplacd (class-last-consumer class)
		   (setf (class-last-consumer class) (list consumer)))
	   (dolist (node (class-nodes class))
	     (when (node-envs node)
	       (enqueue-node node))))))

;;; ******* I bet you this does not handle idnetical antecedent classes ***********
;;; x x -> y will never fire***

;;; Ordering is totally screwey.****

;;; Now it starts to get a little hairy.  Again, this implementation is dumb, but simple.
;;; Someone please give me an efficient version.  This is n^m, I'm sure you
;;; can do nm.  This does a few obvious blocking optimization.

;;; Don't forget that we have to handle the case, that the same node may be in
;;; multiple antecedent classes.

;;; **** should be someway to delete itself from class.
;;; Gets called on (self node supporters).
(defconsumer new-add-conjunctive-class-consumer-internal (node)
	     (classes consumer antecedent-classes)
  (cond ((null classes)) ;*** dequeue
	((not (memq (car classes) (node-classes node))))
	(t (do nil (nil)
	     (setq classes (cdr classes))
	     (cond ((null classes)
		    (add-^-class-consumer-1 consumer antecedent-classes)
		    ;;*** dequeue
		    (return nil))
		   ((null (class-nodes (car classes)))
		    (add-consumer-to-class .consumer. (car classes))
		    (return nil)))))))

(defun new-add-conjunctive-class-consumer (consumer antecedent-classes)
  (unless consumer (error "Another null consumer"))
  (create-consumer NEW-ADD-CONJUNCTIVE-CLASS-CONSUMER-INTERNAL
		   (car antecedent-classes)
		   :CLASSES antecedent-classes
		   :CONSUMER consumer
		   :ANTECEDENT-CLASSES antecedent-classes))

(defconsumer add-conjunctive-class-consumer-1-internal (node) (class ^)
  (dolist (set (node-sets node class (^-nodes ^)))
    (new-add-conjunctive-node-consumer (^-consumer ^) set))
  (when (memq node (cdr class)) (error "consumer invoked twice"))
  (rplacd class (cons node (cdr class))))

;;; See BCW's comments in todo file.  If the same class occurs twice, this
;;; is brain dead.
;;; If the same class occurs twice this is relatively inefficient now, as
;;; it treats each class/position distinct.

(defun add-^-class-consumer-1 (consumer classes &aux ^ slots)
  (unless consumer (error "Null consumer"))
  (dolist (class classes) (push (list class) slots))
  (setq ^ (make-^ :CONSUMER consumer :NODES slots))
  (dolist (class slots)
    (create-consumer ADD-CONJUNCTIVE-CLASS-CONSUMER-1-INTERNAL
		     (car class) :CLASS class :^ ^)))

;;; This returns all possible sets, forcing the membership of node-class argument.
(defun node-sets (node class sets &aux *sets*)
  (node-sets-1 node class sets nil)
  *sets*)

;;; Remember that the same class can occur twice.

(defun node-sets-1 (node class-slot nodes set)
  (cond ((null nodes) (push set *sets*))
	((eq class-slot (car nodes))
	 (node-sets-1 nil nil (cdr nodes) (cons node set)))
	(t (dolist (new-node (cdar nodes))
	     (node-sets-1 node class-slot (cdr nodes) (cons new-node set))))))

;;; A constraint consumer is a consumer that fires if all but one.  It has a
;;; strong built in presumption about the symmetry.  As usual this
;;; handles duplicates at great expense.  This is similar, but far from
;;; identical to the conjunctive consumer.

;;; This also uses exponentially too much space.  Some day fix if matters.  Look at
;;; the summer constraint of old envision to get an idea for how to do it.



;;; ******* if the class types are unique!!************ then much more efficiency is
;;; possible.  
					

;;; **** why can't the closed over variables be stored in the informant.
;;; ***** why isn't a closure a pair?

;;; Consumer will be called (f nodes informant).
;;; This used to use make-art-q-list and g-l-p, which I flushed as there seems
;;; no point in using it as an array.
(defconsumer add-conditional-conjunctive-class-consumer-internal (node) (slot constraint)
  (add-class-consumer-1 .consumer. slot constraint node))

(defun add-conditional-conjunctive-class-consumer
       (consumer classes informant condition &optional test
	&aux constraint count)
;; Many delaying optimizations are possible here.  For example, wait until
;; nodes exist for every class before going to all this work.*****
;; Look at how conditional-class-consumers work****
;; Also delay until condition holds.
  (setq count (length classes)
	classes (mapcar #'list classes)
	constraint (make-constraint :CONSUMER consumer
				    :INFORMANT informant
				    :TEST (or test #'eq)
				    :CONDITION condition
				    :CLASSES classes))
  (dolist (slot classes)
    (create-consumer ADD-CONDITIONAL-CONJUNCTIVE-CLASS-CONSUMER-INTERNAL
		     (car slot)
		     :SLOT slot :CONSTRAINT constraint)))

;;; Need defconsumers for constraints too!!!******
;;; *** the first argument here used to be an ignore, but that caused a bug wh?
(defconsumer add-class-consumer-1-internal (&rest antecedents) (constraint)
  (funcall (constraint-consumer constraint) antecedents (constraint-informant constraint)))
  
;;; The data structures should just be called consumers.****not constraints.
(defun add-class-consumer-1 (me class-slot constraint node &aux informant test sets)
  (setq informant (constraint-informant constraint)
	test (constraint-test constraint))
  (cond ((dolist (j (node-justifications node))
	   (or (funcall test (car j) informant) (return T)))
	 (setq sets (constraint-classes constraint))
	 (rplacd class-slot (cons node (cdr class-slot)))
	 (unless (dolist (slot sets)
		   (cond ((cdr slot))
			 ((eq class-slot slot))
			 (T (return T))))
	   (dolist (set (node-sets node class-slot sets))
	     (if (constraint-condition constraint)
		 (push (constraint-condition constraint) set))
	     ;; Closing over constraint.
	     (create-consumer ADD-CLASS-CONSUMER-1-INTERNAL set :CONSTRAINT constraint))))
	(t (push me (node-justification-consumers node)))))

;;; Consumer is called with (tnode single-unknown-class node-set-with-unknown-absent)
;;; The problem solver decides what to do about this.

;;; *** this could be haired up as the conjunctive one.
(defun add-constraint-class-consumer (ignore consumer2 classes informant &optional test) 
  (add-general-constraint consumer2 classes informant test nil))

;;; This constraint is only active if it is conjoined with a node.
;;; ****** condition
;;; ***** add consumer to condition to wait until it gets in.*******
;;; ***** this is rather inefficient.
;;; All these constraint consumers suffer from the problem that the same class may
;;; appear twice.******
;;; Could optimize case where each class occurs once.
;;; Could optimize case where a node is in at most one class.

(defun add-c-c-c-c (consumer classes informant condition &optional test)
  (add-general-constraint consumer classes informant test condition))

(defconsumer add-general-constraint-internal (node) (slot constraint)
  (add-constraint-class-consumer-1 .consumer. slot constraint node))

(defun add-general-constraint (consumer classes informant test condition
			       &aux constraint count)
  (setq count (length classes)
	classes (mapcar #'list classes)
	constraint (make-constraint :CONSUMER consumer
				    :INFORMANT informant
				    :TEST (or test #'eq)
				    :CONDITION condition
				    :CLASSES classes))
  (dolist (slot classes)
    (create-consumer ADD-GENERAL-CONSTRAINT-INTERNAL (car slot)
		     :SLOT slot :CONSTRAINT constraint)))

;;; Remember this gets called when the node is added to the class which is part of
;;; a class constraint.
;;; *** cold also check that the cosntraint-informant is also in......
;;; Look at old envision for efficiency hacks here.****
(defun add-constraint-class-consumer-1 (me class-slot constraint node
					&aux unknown-class informant test)
  (setq informant (constraint-informant constraint)
	test (constraint-test constraint))
  (cond ((dolist (j (node-justifications node))
	   (or (funcall test (car j) informant) (return T)))
	 (unless (dolist (slot (constraint-classes constraint))
		   (cond ((cdr slot))
			 ((eq class-slot slot))
			 (unknown-class (return T))
			 (t (setq unknown-class slot))))
;	   (format T "~% Adding ~A to ~A for ~A" node class-slot constraint)
	   (if unknown-class
	       (dolist (set (constraint-node-sets node class-slot constraint))
		 ;;; ************ save conses, put on end.
		 (add-constraint-consumer (car unknown-class) constraint set))
	       (dolist (unknown-class (constraint-classes constraint))
		 (unless (eq unknown-class class-slot)
		   (dolist (set (node-setsa node class-slot unknown-class constraint))
		     (add-constraint-consumer (car unknown-class) constraint set))))))
	 (rplacd class-slot (cons node (cdr class-slot))))
	(t (push me (node-justification-consumers node)))))


;;; **constraints need to be consumers.
(defconsumer add-constraint-consumer-internal (&rest antecedents) (constraint unknown-class)
  (funcall (constraint-consumer constraint)
	   nil
	   unknown-class
	   antecedents
	   (constraint-informant constraint)))

(defun add-constraint-consumer (unknown-class constraint set)
  ;;; ************ save conses, put on end.
  (if (constraint-condition constraint) (push (constraint-condition constraint) set))
  (create-consumer ADD-CONSTRAINT-CONSUMER-INTERNAL
		   set :CONSTRAINT constraint :UNKNOWN-CLASS unknown-class))

;;; Like other node sets but it forcibly excludes ignore-class.
(defun node-setsa (node class-slot ignore-class-slot constraint &aux *sets*)
  (node-setsa-1 node class-slot ignore-class-slot
		(constraint-classes constraint) nil)
;  (format T "~% Computed ~D node sets" (length *sets*))
  *sets*)

(defun node-setsa-1 (node class-slot ignore-class-slot nodes set)
  (cond ((null nodes) (push set *sets*))
	((eq class-slot (car nodes))
	 (node-setsa-1 nil nil ignore-class-slot (cdr nodes) (cons node set)))
	((eq ignore-class-slot (car nodes))
	 (node-setsa-1 node class-slot nil (cdr nodes) set))
	(t (dolist (new-node (cdar nodes))
	     (node-setsa-1 node class-slot ignore-class-slot
			   (cdr nodes) (cons new-node set))))))

;;; *install this into above functin **
(defun justified-by (node informant me)
  (unless (dolist (j (node-justifications node)) (or (eq (car j) informant) (return T)))
    (push me (node-justification-consumers node))
    T))

(defun constraint-node-sets (node class-slot constraint &aux *sets*)
  (constraint-node-sets-1 node class-slot (constraint-classes constraint) nil)
  *sets*)

(defun constraint-node-sets-1 (node class-slot nodes set)
  (cond ((null nodes) (push set *sets*))
	((eq class-slot (car nodes))
	 (constraint-node-sets-1 nil nil (cdr nodes) (cons node set)))
	((cdar nodes)
	 (dolist (new-node (cdar nodes))
	   (constraint-node-sets-1 node class-slot (cdr nodes) (cons new-node set))))
	(t (constraint-node-sets-1 node class-slot (cdr nodes) set))))

;;; A disjunctive constraint consumer, is a kind of constraint which gets triggered 
;;; when any one(!?) of its inputs gets a value, however, reflections are suppressed.

;;; This consumer gets called with [node informant].

(defun add-disjunctive-constraint-class-consumer (consumer classes informant
						  &aux class-consumer)
  (setq class-consumer (let-closed ((consumer consumer)
				    (informant informant))
			 #'(lambda (me node)
			     (if (dolist (j (node-justifications node))
				   (or (equal (car j) informant)
				       (return T)))
				 (funcall consumer node informant)
				 (push me (node-justification-consumers node))))))
  (dolist (class classes) (add-consumer-to-class class-consumer class)))

;;; A conditional disjunctive constraint consumer, is a kind of constraint which gets
;;; get triggered when any one of its inputs gets a value AND the node it is
;;; conditional on holds.

;;; **************** make all callers to add-d-c-c-c take the two arguments class/condition.
;;;********* consumer is not called this way any more.
(defconsumer add-c-d-c-c-c-internal (&rest supporters) (consumer informant)
  (funcall consumer supporters informant))

;;; This consumer gets called with [(class-node condition-node) informant]
;;;add-conditional-disjunctive-constraint-class-consumer
(defun add-c-d-c-c-c (consumer classes informant condition)
  (add-disjunctive-constraint-class-consumer
    (let-closed ((condition condition)
		 (consumer consumer))
      #'(lambda (class-node informant)
	  (create-consumer ADD-C-D-C-C-C-INTERNAL
			   (list class-node condition)
			   :CONSUMER consumer :INFORMANT informant)))
    classes
    informant))


;;; The new agenda mechanism is organized by the environment waiting to be run.
;;; Call this when a consumer first gets attached to a node, or when a new environment
;;; is added to a node.  :AGENDA/:BLOCKED indicates which queue this node is currently on.

;;; This enqueues the node onto the current agenda or blocked nodes database.
(defun enqueue-node (node)
  (selectq (node-enqueued? node)
    ;; This used to call (insert-in-agenda node), but that should be redundant.
    (:AGENDA)
    (:BLOCKED (unless (fast-simple-blocked? node nil)
;;; Should be some other flag.  Fast-simple-blocked? will remove this node
;;; from the queues if it just became unblocked.
		(insert-in-agenda node)))
    (T (or (blocked? node)
	   (insert-in-agenda node)))))

;;; This is FIFO as we want. *agenda* consists of triples: (size last . envs)
;;; This does not check for blocking, however, I suspect that some of the functions
;;; that call this should because they might inadvertently schedule a consumer
;;; that should not be run.


(defun insert-in-agenda (node &aux env size consumer-nodes nodes envs)
  (setf (node-enqueued? node) :AGENDA)
  (setq env (shortest-env (node-envs node))
	consumer-nodes (env-consumer-nodes env))
  ;; If the env has consumers, then it already is on agenda.
  ;; Consumer-nodes should not have duplicates on it.
  (cond (consumer-nodes (setq nodes (ncons node))
			(rplacd (car consumer-nodes) nodes)
			(rplaca consumer-nodes nodes)
			)
	(t (setq nodes (ncons node))
	   (setf (env-consumer-nodes env) (cons nodes nodes))
	   (setq size (env-count env)
		 envs (ncons env))
	   (do ((previous nil next)
		(next *agenda* (cdr next)))
	       ((null next)
		(if previous (rplacd previous (ncons (cons size (cons envs envs))))
		    (setq *agenda* (ncons (cons size (cons envs envs))))))
	     (cond ((= size (caar next))
		    (rplacd (cadar next) envs)
		    (rplaca (cdar next) envs)
		    (return nil))
		   ((> (caar next) size)
		    (if previous
			(rplacd previous (cons (cons size (cons envs envs)) next))
			(setq *agenda* (cons (cons size (cons envs envs)) next)))
		    (return nil)))))))

;;; It might be worth organizing labels so that the shortest environment was always
;;; first.
(defun shortest-env (envs &aux size new-size new-env)
  (dolist (env envs)
    (setq new-size (env-count env))
    (if (or (null size) (< new-size size)) (setq size new-size new-env env)))
  new-env)

;;; We'll find these when we pop them off.
(defun dequeue-node (node)
  (setf (node-enqueued? node) nil))

;;; Remember a consumer is always a list of lists.
(defun run (&aux node consumer start-time consumer-nodes envs)
  (setq start-time (get-internal-run-time))
  (do nil ((null *agenda*))
    (when (and *current-context* *current-nogoods*)
      (dependency-directed-backtracking))
    (when (eq *current-context* 'FAIL)
      (setq *agenda* nil)
      (return nil))
    (setq envs (cddar *agenda*))
    (cond ((null envs) (setq *agenda* (cdr *agenda*)))
	  ((null (setq consumer-nodes (env-consumer-nodes (car envs))))
	   (cond ((null (cdr envs)) (setq *agenda* (cdr *agenda*)))
		 (t (rplacd (cdar *agenda*) (cdr envs)))))
;;**** this is not true anymore. We have to check for blocked-envs all the time because
;; nothing updates their queuing status when current context changes.******* FIX.
;; so I uncommented out the commented out block-node?
;	  ((blocked-env? (car envs)))
	  (t (setq node (cadr consumer-nodes))
	     (rplacd consumer-nodes (cddr consumer-nodes))
	     (unless (cdr consumer-nodes)
	       (setq consumer-nodes nil)
	       (setf (env-consumer-nodes (car envs)) nil))
	     (cond ((neq :AGENDA (node-enqueued? node)))
		   ((i-out? node) (dequeue-node node))
		   ;; This line ***must be here***
		   ((blocked? node)) ;;;******?????Don't when inserted.
		   ((not (memq (car envs) (node-envs node))) (insert-in-agenda node))
		   (t (if *trace*
			  (format T "~% Running consumers of ~A : ~A" node
				  (node-consumers node)))
		      (dequeue-node node)
		      ;;**** isn't this wrong as a shorter environment may be entered???
		      ;; Every slot on consumers points to one before.  I flushed locatives.
		      (do nil (nil)
			(unless (node-envs node) (return nil))
			(if (do ((consumers (node-consumers node) (cdr consumers)))
				((null consumers) T)
			      (setq consumer (cdar consumers))
			      (when consumer
				(rplaca consumers consumer)
				(invoke-consumer (car consumer) node)
				(process-queued-nogoods)
				(return nil)))
			    (return nil))
			(and *current-context* *current-nogoods* (return nil))))))))
  ;(format T "~% Running time is:~D seconds" (time-taken start-time))
 )

(defun invoke-consumer (consumer *node*)
  (incf *consumer-invokations*)
  (cond ((or (functionp consumer) (symbolp consumer)) (funcall consumer consumer *node*))
	(t (funcall (consumer-function consumer) consumer *node*))))


;;; Worked out by Johan de Kleer and Brian Williams.

(defun make-control-disjunction-node (assumptions)
  assumptions
  )  

;;; This is the stuff I promised not to do: dependency-directed backtracking.
;;; This version is simplistic:
;;; o Assumes each assumption 

(defvar *trace-ddb* nil)

;;; Control stack entries are now: (assumptions current-assumption previous-environment).

(defun assert-control-disjunction (assumptions)
  (dolist (assumption assumptions)
    (setf (assumption-addb-index assumption) *addb-count*)
    (push assumptions (assumption-disjunctions assumption)))
  (push (list assumptions (car assumptions) (or *current-context* *empty-env*))
	*control-stack*)
  (if *trace-ddb* (format T "~% Adding first assumption of a new control disjunction: ~A"
			  (string-assumption (car assumptions))))
  (setq *current-context*
	(fast-cons-env (or *current-context* *empty-env*) (car assumptions) T))
  (incf *addb-check-count*)
  (unblock)
  ;; We really need to check whether anything is blocked now.*****
  )

(defvar *show-backtracks* nil)

;;; Assumes you don't like the current context anymore (it need not be nogood).
;;; Returns NIL on failure, in which case *current-context* is the atom FAIL.
(defun dependency-directed-backtracking (&aux oldest-contra next-current-context)
  (if (eq *current-context* 'FAIL)
      (return-from DEPENDENCY-DIRECTED-BACKTRACKING nil))
  ;; First search back to a point which avoids all known contradictions.
  (setq oldest-contra (unwind-stack))
  ;; At this point oldest-contra points to the current stack position which is failing.
  ;; Here we've backtracked to the last contradiction, or else there is an ATMS bug somewhere.
  ;; We've backtracked to the last contradiction, pick a next assumption if we can.
  (setq next-current-context nil)
  (do nil (nil)
    (unless oldest-contra (return))
    (when (dolist (next-assumption (cdr (memq (cadar oldest-contra) (caar oldest-contra))))
	    ;; If all are false, this should do more intelligent things.*****
	    (incf *addb-check-count*)
	    (cond ((i-false? next-assumption))
		  ((and (not (simple-envp (setq next-current-context
						(fast-cons-env *current-context*
							       next-assumption T))))
			(env-contradictory next-current-context))
		   (if *trace-ddb* (format T "~% Can't add next assumption: ~A"
					   (string-assumption next-assumption))))
		  (t (if *trace-ddb* (format T "~% Adding next assumption: ~A"
					     (string-assumption next-assumption)))
		     (rplaca (cdar oldest-contra) next-assumption)
		     (rplaca (cddar oldest-contra) *current-context*)
		     (setq *current-context* next-current-context)
		     (if (complete oldest-contra) (return T))
		     (setq *current-context* (caddar oldest-contra)))))
      (if (or *trace-ddb* *show-backtracks*)
	  (format T "~% Backtracked to: ~A" *current-context*))
      (unblock)
      (return-from DEPENDENCY-DIRECTED-BACKTRACKING T))
    (if (and (or (eql *h4* 1) *resolve-by-labeling* *resolve-by-ordered-labeling*)
	     *contradict-solutions*)
	(error "Can't happen"))
    (incf *exhaustions*)
    (if *trace-ddb* (format T "~% Exhausted the disjunction: ~A" (caar oldest-contra)))
    (setq oldest-contra (cdr oldest-contra)
	  *current-context* (caddar oldest-contra)))
  (if *trace-ddb*
      (format T "~%Can't backtrack anymore --- no more consistent interpretations."))
  (setq *current-context* 'FAIL)
  nil)

;;; This only bothers remembering those nogoods which are subsets of the current
(defun signal-nogood (env)
  (and *current-context*
       (vector-subset (env-vector env) (generic-env-vector *current-context*))
       (push env *current-nogoods*)))

(defun backtrack (print)
  (dotimes (j 1000000.)
    (if print (format T "~% Solution ~D : ~A, Environments = ~D"
		      j (string-env *current-context*) *env-counter*))
    
    (if (and *contradict-solutions*
	     (or (eql *h4* 1) *resolve-by-labeling* *resolve-by-ordered-labeling*))
	(let (*going-nodes*)
	  (contradictory-env (simple-to-env *current-context*) '(SPECIOUSNESS))
	  (process-changed-nodes)
	  ))
    (process-queued-nogoods)
    (unless (dependency-directed-backtracking) (return nil))
    (if (eq 'FAIL (catch 'CONTRADICTION (run))) (return nil))
    (if (eq *current-context* 'FAIL) (return nil))
    ))
  
;;; Simply unwind the stack until one element of every *current-nogoods* is removed.
;;; Leaves *current-context* pointing at a consistent context.  oldest-contra at
;;; the next stack position.  This assumes every nogood is a subset of the current
;;; context.  If not this will either error out or produce bad results.

(defun unwind-stack (&aux stack previous-stack)
  (setq stack *control-stack*)
  (if *trace-ddb*
      (format T "~% Backtracking because current context contains nogoods: ~A"
	      *current-nogoods*))
  (if *current-nogoods*
      (dolist (nogood *current-nogoods*)
	(do nil (nil)
	  (cond ((vector-subset (env-vector nogood) (generic-env-vector *current-context*))
		 (if *trace-ddb*
		     (format T "~% To avoid contradiction ~A, removing the assumption: ~A"
			     (string-env nogood) (string-assumption (cadar stack))))
; We should delete the singleton nogoods soon...
;		 (when (eq nogood (assumption-env (cadar stack)))
;		   (rplaca (car stack) (fdelq (cadar stack)
		 
		 (setq *current-context* (caddar stack)
		       previous-stack stack
		       stack (cdr stack))
		 )
		(t (return nil)))))
	 
      (setq *current-context* (caddar stack)
	    previous-stack stack))
  (setq *current-nogoods* nil)
  previous-stack)

;;; oldest-contra is points at the current point in the stack.

(defun complete (oldest-contra &aux new-control-stack current-context)
  (setq current-context *current-context*)
  (cond ((eq oldest-contra *control-stack*))
	(t (do ((cs *control-stack* (cdr cs)))
	       ((eq cs oldest-contra))
	     (setq new-control-stack cs))
	   ;; *** the current context should be doubly linked.
	   (dolist (assumption (caar new-control-stack))
	     (incf *addb-check-count*)
	     (cond ((i-false? assumption)
		    ;; Something should delete this assumption.  But not here
		    )
		   ((and (not (simple-envp (setq *current-context*
						 (fast-cons-env *current-context*
								assumption T))))
			 (env-contradictory *current-context*))
		    (setq *current-context* current-context))
		   (t (rplaca (cdar new-control-stack) assumption)
		      (rplaca (cddar new-control-stack) current-context)
		      (cond ((complete new-control-stack) (return T))
			    (t (setq *current-context* current-context)))))))))

;;; There must be one environment whose disjuntion assumptions are all in current context.
;;; **** probably a flag necessry.
;;; **** also a node can become ublocked when an environment is added to it.
;;; **** this works but is inefficient has hell.
;;; **** also a node can be put here more than once.  (May not matter).
;;; Note that the node is not considered queued after it is blocked.
(defun blocked? (node)
  (when *current-context*
    (cond ((eq (node-enqueued? node) :BLOCKED))
	  ((not (fast-simple-blocked? node)) nil)
	  (t (setf (node-enqueued? node) :BLOCKED) T))))

;;; This simply checks whether the given node is blocked
;;; BUG HERE? consistent with.*********

;;; Obsolete?****

(defun simple-blocked? (node &aux vector)
  (when *current-context*
    (setq vector (generic-env-vector *current-context*))
    (not (dolist (e (node-envs node))
	   (unless (dolist (c (env-assumptions e))
		     (and (assumption-disjunctions c)
			  (not (vector-member c vector))
			  (return T)))
	     (return T))))))

;;; This uses the first assumption as an index, if there are multiple environments
;;; with the same first index this does extra work.  If this conses a lot
;;; we can save those too.  If the node is now unblocked, this makes sure
;;; the node is removed from all blocked-node lists.  Notice we can still
;;; sometimes get duplicates because envs get added and removed as problem solving
;;; progresses.  Nothing takes the blocking off when an environment is removed from a node.
;;; Maybe we should???

;;; The idea is that every environment of the label of a node should be blocked waiting
;;; for exactly one assumption.  If needed-assumption is NIL, clean up the data structures.
(defun fast-simple-blocked? (node &optional needed-assumption &aux vector blocked)
  (cond ((null *current-context*) nil)
	((i-true? node) nil)
	(T (setq vector (generic-env-vector *current-context*))
	   (dolist (e (node-envs node))
	     (when (or (null needed-assumption)
		       (memq needed-assumption (env-assumptions e)))
	       (unless (dolist (c (env-assumptions e))
			 (when (and (assumption-disjunctions c) (not (vector-member c vector)))
			   (pushnew c blocked)
			   (return T)))
		 ;; This is slightly overkill because the context might become inconsistent
		 ;; This unblocks this node from all possible places.
		 (unless needed-assumption
		   (dolist (e (node-envs node))
		     (dolist (a (env-assumptions e))
		       (setf (assumption-blocked-nodes a)
			     (fdelq1 node (assumption-blocked-nodes a))))))
		 (return-from FAST-SIMPLE-BLOCKED? nil))))
	   ;; It may not be worth the trouble detecting duplicates...
	   (dolist (b blocked) (pushnew node (assumption-blocked-nodes b)))
	   T)))

;;; This is non-standard actually: A consumer is run only if it has some 
;;; environment whose every control assumption appears in the current context.
;;; We can recycle conses here if we need to.  Or have an extra field in a consumer/node
;;; for threading.
(defun unblock ()
  (dolist (a (generic-env-assumptions *current-context*))
    (dolist (b (assumption-blocked-nodes a))
      (cond ((neq (node-enqueued? b) :BLOCKED))
	    ((fast-simple-blocked? b a))
	    (T (insert-in-agenda b))))
    (setf (assumption-blocked-nodes a) nil)))
  
;;; New blocking code for run.

;;; An environment is blocked if all its control disjunctions don't appear in the
;;; current context.  ***[unimplemented] it is consistent with *current-context*

;;; Envs point to nodes, so the following when implemented is much better.


(defun blocked-env? (env)
  (when (simple-blocked-env? env)
    (push env *blocked-envs*)
    (dolist (node (env-consumer-nodes env))
      (enqueue-node-if-not-blocked node env))
    T))

(defun simple-blocked-env? (env)
  (dolist (a (env-assumptions env))
    (and (assumption-disjunctions a)
	 (not (memq a (generic-env-assumptions *current-context*)))
	 (return T))))

;(defun enqueue-node-if-not-blocked (node known-blocked-env)
;  known-blocked-env
  

;;; Finds the shortest environment among envs which is not blocked.
(defun shortest-non-blocked-env (envs &aux size new-size new-env)
  (dolist (env envs)
    (setq new-size (env-count env))
    (and (or (null size) (< new-size size))
	 (not (simple-blocked-env? env))
	 (setq size new-size new-env env)))
  new-env)