;;; -*- Syntax:Common-Lisp; Mode: LISP; Package: (TMS Lisp 1000.); Base: 10. -*-

(in-package 'tms)

"(c) Copyright 1986, 1987, 1988 Xerox Corporation.  All rights reserved.  Subject to
the following conditions, permission is granted to use and copy this
software and to prepare derivative works:  Such use, copying or
preparation of derivative works must be for non-commercial research or
educational purposes; each copy or derivative work must include this
copyright notice in full; a copy of each completed derivative work must
be returned to:  DEKLEER@XEROX.COM (Arpanet) or Johan de Kleer,
Xerox PARC, 3333 Coyote Hill Road, Palo Alto, CA 94304.  This software
is made available AS IS, and Xerox Corporation makes no warranty about
the software or its performance."

;;; Simple debugging functions.
(defun print-nodes ()
  (format T "~% Nodes:")
  (dolist (node *nodes*)
    (print (list node
		 (string-envs (node-envs node))
		 (node-enqueued? node)
		 (node-consumers node)
		 ))))

;;; This summarizes all the ATMS inputs.
(defun print-inputs ()
  (format T "~%ATMS input forms:")
  (dolist (node *nodes*)
    (dolist (j (node-justifications node)) (print-justification node j)))
  (dotimes (i 99.) (dolist (or (aref *ors* i)) (print-or or)))
  )

(defun print-or (or)
;  (format T "~%choose {")
  (do ((disjuncts (disjunction-disjuncts or) (cdr disjuncts)))
      ((null disjuncts))
    (princ (assumption-string (car disjuncts)))
    (if (cdr disjuncts) (princ ",")))
;  (princ "}")
  )

;;; Prints out in format of papers, uses node-datums, uses [] to indicate an assumption.
(defun print-justification (node just)
  (terpri)
  (do ((a (cdr just) (cdr a)))
      ((null a))
    (princ (n-a-string (car a)))
    (if (cdr a) (format T ",")))
  (princ "=>")
  (princ (n-a-string node))
  (princ "      from ")
  (princ (car just)))

(defun n-a-string (n-a)
  (cond ((eq n-a *contra-node*) "FALSE")
	((assumption? n-a) (format nil "[~A]" (n-a-datum n-a)))
	(T (format nil "~A" (n-a-datum n-a)))))

(defun print-contradictions (&aux (count 0))
  (walk-contradictions 
    #'(lambda (env)
	(incf count)
	(format T "~&~A (count: ~D)~%   (~A)"
		(string-env env) (env-count env) (env-contradictory env))))
  (format T "~% ~D contradictions total." count))

(defun check-contradictions ()
  (walk-contradictions #'(lambda (nogood)
			   (if (subsumed-nogood? nogood)
			       (error "Contradiction error")))))

(defun print-ors ()
  (dotimes (i 100.) (mapc #'print-or (aref *ors* i))))

(defun test-general-weave (known-envs new-partial-envs consequent
		       &optional (node-to-ignore 'IGNORE) &aux k1 k2 n1 n2)
  (multiple-value-setq (k1 n1)
    (ogeneral-weave known-envs new-partial-envs consequent node-to-ignore))
  (multiple-value-setq (k2 n2)
    (user-general-weave known-envs new-partial-envs consequent node-to-ignore))
  (or (= (length k1) (length k2)) (error "Known mismatch"))
  (or (= (length n1) (length n2)) (error "New mismatch."))
  (values k2 n2))

(defvar *node-to-ignore* nil)			; Node to ignore in weaving.
(defvar *known-envs*)
(defvar *new-envs*)

(defun ogeneral-weave (*known-envs* new-partial-envs consequent
		      &optional (*node-to-ignore* 'IGNORE)
		      &aux *new-envs*)
  (unless (dolist (n-a consequent) (if (i-false? n-a) (return T)))
    (dolist (new-partial-env new-partial-envs) (oweave consequent new-partial-env))
    (when *new-envs*
      (setq *known-envs* (fcopylist *known-envs*))
      (dolist (new-env *new-envs*)
	(map #'(lambda (known-env)
		 (and (car known-env)
		      (subset-env? new-env (car known-env))
		      (rplaca known-env nil)))
	     *known-envs*))
      (setq *known-envs* (nconc (fdelqa nil *known-envs*) *new-envs*))))
  (values *known-envs* *new-envs*))


;;; This is slow but right.
(defun oweave (nodes-to-process partial-env &aux new-env n-a)
  (setq n-a (car nodes-to-process))
  (cond ((or (eq n-a *node-to-ignore*) (if n-a (i-true? n-a)))
	 (oweave (cdr nodes-to-process) partial-env))
	((dolist (known-env *known-envs*)
	   (if (subset-env? known-env partial-env) (return T))))
	((dolist (new-env *new-envs*)
	   (if (subset-env? new-env partial-env) (return T))))
	((null nodes-to-process)
	 (map #'(lambda (new-env)
		  (when (subset-env? partial-env (car new-env))
		    (rplaca new-env nil)))
	      *new-envs*)
						;*** too many delqs'
	 (setq *new-envs* (cons partial-env (fdelqa nil *new-envs*))))
	((node? n-a)
	 (dolist (env (node-envs n-a))
	   (unless (env-contradictory (setq new-env (union-env partial-env env)))
	     (oweave (cdr nodes-to-process) new-env))))
	(t (unless (env-contradictory (setq new-env (cons-env partial-env n-a)))
	     (oweave (cdr nodes-to-process) new-env)))))



;;; This deletes any supersets of set2 in set1 preserving as much list structure as possible.
;;; Is hackery really worth it?
;(defun safely-delete-subsets (set1 set2 &aux new-set2)
;  (when set2
;    (setq new-set2 (safely-delete-subsets set1 (cdr set2)))
;    (cond ((dolist (item1 set1) (if (subset-env? item1 (car set2)) (return T))) new-set2)
;	  ((eq (cdr set2) new-set2) set2)
;	  (t (cons (car set2) new-set2)))))




(defun print-statistics (solutions &aux count)
  (format T "~% Number of solutions: ~D" (length solutions))
  (format T "~% Number of consumers invoked ~D" *consumer-invokations*)
  (format T "~% Number of TMS nodes: ~D" (1- *node-counter*))
  (format T "~% Number of TMS classes: ~D" (length *classes*))
  (format T "~% Number of contradictions: ~D" (1- *contra-counter*))
  (format T "~% Number of assumptions: ~D" (1- *assumption-counter*))
  (format T "~% Number of environments: ~D" (1- *env-counter*))
  (format T "~% Number of large union calculations: ~D" *big-union-count*)
  (format T "~% Number of full resolution steps: ~D" *full-resolution-count*)
  (format T "~% Minimal nogood resolutions ~D" *minimal-nogood-count*)
  (format T "~% Number of failed full resolution steps: ~D" *full-resolution-fail-count*)
  (format T "~% Number of justifications: ~D" *justification-count*)
  (format T "~% Number of binary resolutions: ~D" *binary-resolution-count*)
  (format T "~% Number of exhaustions: ~D" *exhaustions*)
  (format T "~% Number of edges in ADDB search tree: ~D" *addb-check-count*)
  (format T "~% Size of maximum nogood constructed: ~D" *max-contra-count*)
  (format T "~% Size of maximum nogood resolved with: ~D" *max-nogood-used-size*)
  (dotimes (i (1+ *max-contra-count*))
    (setq count (count-nogoods i))
    (when (> count 0)
      (format T "~% There are ~D base contradictions of size ~D" count i)))
  (dotimes (i (array-length *environments*))
    (when (aref *environments* i)
      (format T "~% There are ~D base environments of size ~D"
	      (length (aref *environments* i)) i)))
  (justification-statistics)
  solutions)

(defun count-nogoods (size)
  (let (count)
    (setq count 0)
    (walk-tree #'(lambda (ignore) (incf count)) (aref *nogood-trees* size))
    count))

;;; This gives info about free nodes too --- needs more work.
(defun justification-statistics (&aux j-array c-array too-big size
				 (true 0) (false 0) (nothing 0))
  (setq j-array (make-array 1000. :ELEMENT-TYPE '(INTEGER 0 65536) :INITIAL-ELEMENT 0)
	c-array (make-array 1000. :ELEMENT-TYPE '(INTEGER 0 65536) :INITIAL-ELEMENT 0))
  (dolist (n *nodes*)
    (if (i-true? n)
	(incf true)
	(if (i-false? n) (incf false) (unless (n-a-envs n) (incf nothing))))
    (setq size (length (node-justifications n)))
    (if (> size 1000.)
	(push n too-big)
	(incf (aref j-array size)))
    (setq size (length (node-consequents n)))
    (if (> size 1000.)
	(push n too-big)
	(incf (aref c-array size))))
  (format T "~% ~D true nodes; ~D false nodes; ~D empty labels" true false nothing)
  (format T "~% ~D nodes had too many justifications/consequents" (length too-big))
  (format T "~% Number of justifications per node:")
  (dotimes (i 1000.)
    (setq size (aref j-array i))
    (unless (= size 0)
      (format T "~%  ~D nodes with ~D justifications" size i)))
  (format T "~% Number of consequences per node:")
  (dotimes (i 1000.)
    (setq size (aref c-array i))
    (unless (= size 0)
      (format T "~%  ~D nodes with ~D consequences" size i))))

(defun check-order (as &optional msg)
  (do ((as as (cdr as)))
      ((null (cdr as)))
    (unless (assumption-orderp (car as) (cadr as)) (error "Illegal order"))))

(defun check-env-array ()
  (dotimes (i (array-length *environments*))
    (dolist (e (aref *environments* i))
      (if (env-contradictory e)
	  (format T "~% Contradictory env ~A found in slot ~D" e i)))))

(defun test-cons-env (l) (dolist (r l) (cons-env (car r) (cadr r))))

(defun clear-cons-env-cache ()
  (dotimes (i (array-length *environments*))
    (dolist (e (aref *environments* i))
      (setf (env-cons-env-cache e) nil))))

(defun clear-cons-env-cache1 ()
  (dotimes (i (array-length *environments*))
    (dolist (e (aref *environments* i))
      (if (env-cons-env-cache e)
	  (rplacd (env-cons-env-cache e) nil)))))

(defun lookat-caches ()
  (dotimes (i (array-length *environments*))
    (dolist (e (aref *environments* i))
      (when (cdr (env-cons-env-cache e))
	(format T "~% ~A ~A" e (cdr (env-cons-env-cache e)))))))

;;; ******** need a list of environments somehwerew********** *environments* is not it.
(defun check-env-cache (&aux cache count acount)
  (dotimes (i (array-length *environments*))
    (dolist (e (aref *environments* i))
      (when (setq cache (env-cons-env-cache e))
	(setq count 0 acount 0)
	(dolist (c (cdr cache)) (if (assumption-contradictory (car c)) (incf acount))
				(if (env-contradictory (cdr c)) (incf count)))
	(if (> count 0) (format T "~% ~A has ~D contradictions in cache" e count))
	(if (> acount 0) (format T "~% ~A has ~D bad assumptions in cache" e acount))))))

(defun clear-string-cache ()
  (dotimes (i (array-length *environments*))
    (dolist (e (aref *environments* i))
      (setf (env-string e) nil))))
						
;;; This finds how many environments are in the hash table and how many
;;; are contradictory.
#+(OR :ZL :CL-ZL)
(defun scan-envs (&aux goods vectors contras slot)
  (setq goods 0 vectors 0 contras 0)
  (let ((ptr (hash-array *env-hash-table*)))
    (declare (sys:array-register ptr ))
    (dotimes (p3 (hash-p2 *env-hash-table*))
      (setq slot (aref ptr p3))
      (cond ((null slot))
	    ((listp slot) (incf vectors))
	    ((env-contradictory slot) (incf contras))
	    (t (incf goods)))))
  (format T "~% Of ~D possible, ~D are good, ~D are vectors, ~D are contradictory"
	  (hash-p2 *env-hash-table*) goods vectors contras))

;;; This is only partially right, because it ignores earlier nogoods (if there are
;;; more than three nodes).
(defun consistent-nodes? (the-nodes)
  (do ((envs (tms::node-envs (car the-nodes))
	     (tms::union-envss (tms::node-envs (car nodes)) envs))	;
       (nodes (cdr the-nodes) (cdr nodes))
       (last-envs nil)
       (last-node (car the-nodes)))
      ((or (null envs) (null nodes))
       (cond ((null envs)
	      (values nil (find-env-union-killer (tms::node-envs last-node)
							     last-envs)))
	     ((null (cdr nodes)) ;out of them
	      envs)))
    (setq last-envs envs last-node (car nodes))))

;;; This is only partially right, because it only returns the first nogood.
(defun find-env-union-killer (e1s e2s &aux result)
  (dolist (e1 e1s)
    (dolist (e2 e2s)
      (setq result (tms::find-nogood-for-assumptions
		     (tms::assumptions-union (tms::env-assumptions e1)
					     (tms::env-assumptions e2))
		     nil))
      (if result (return-from FIND-ENV-UNION-KILLER result)))))

(defun print-union-stats (&aux total max count contras l bige)
  (setq total 0 contras 0 count 0)
  (do ((i 1 (1+ i)))
      ((> i *max-env-count*))
    (dolist (e (aref *environments* i))
      (incf count)
      (setq l (or (car (env-unions e)) 0))
      (incf total l)
      (when (or (null max) (< max l)) (setq max l bige e))
      (dolist (p (cdr (env-unions e)))
	(if (eq (cdr p) *contra-env*) (incf contras)))))
  (format T "~% ~D environments, longest cache= ~D(~A), average cache=~D, % contra=~D"
	  count max bige (// total count) (// (* contras 100.) total)))

(defun print-hs ()
  (selectq *h4*
    (NIL (format T "~% Rule *h4* is disabled."))
    (1 (format T "~% Rule *h4* is enabled."))
    (2 (format T "~% Rule *h4* applies to other than binary disjunctions."))
    (T (format T "~% Illegal value for *h4*: ~A" *h4*)))
  (if *h45* (format T "~% Rule *h45* is enabled.") (format T "~% Rule *h45* is disabled.")))


(defun why-nogood (n-a-e)
  (cond ((assumption? n-a-e)
	 (why-nogood-assumption n-a-e))))

(defun why-nogood-assumption (a)
  (selectq (car (assumption-contradictory a))
    (BASE (format T "~% Assumption ~A is false because its singleton environment is.")
	  (why-nogood-env (cdr (assumption-contradictory a))))
    (T (error "Uninimplemented"))))

;;; Returns a justification/resolution for false killing the nogood.
(defun why-nogood-env (env &aux reason)
  (unless (env-contradictory env) (error "Environment is consistent."))
  (if (eq env *contra-env*)
      (error "Environment is non-unique --- try why-nogood-assumptions."))
  (setq env (find-base-contradiction env))
  (setq reason (cdr (env-contradictory env)))
  (selectq (car reason)
    (BACK-PROPAGATE (cadr reason))
    (LABEL-FOR-FALSE2 (fourth reason))
    (LABEL-FOR-FALSE (cddr reason))
    (RESOLUTION (cdr reason))
    (WEAVING-FOR-FALSE (cdr reason))
    (T (error "Unknown contradiction reason"))))

;;; Finds the env killing the assumptions, or in a few cases a justification
;;; which is killing it (i.e., the env itself was never created).
(defun find-nogood-for-assumptions (assumptions &optional (is-nogood T) &aux env v)
  (dolist (a assumptions)
    (if (i-false? a) (return-from FIND-NOGOOD-FOR-ASSUMPTIONS (assumption-env a))))
  (setq env (find-or-make-env assumptions))
  (cond ((eq env *contra-env*)
	 (cond ((not (compatible-assumptions? assumptions
					      (setq v (make-env-vector* assumptions))))
		(do ((as assumptions (cdr as)))
		    ((null as) (error "Can't happen"))
		  (dolist (a (cdr as))
		    (when (vector-member (car as) (assumption-binary-vector a))
		      (return-from FIND-NOGOOD-FOR-ASSUMPTIONS
			(or (double-if-exists (car as) a)
			    (list *contra-node* '<UNKNOWN-INFORMANT> (car as) a)))))))
	       ((new-contradictory-env-assumptions? assumptions
						    (length assumptions)))
	       (t (error "Can't find the nogood killing the assumptions, but something is"))))
	((env-contradictory env) (find-base-contradiction env))
	(is-nogood (error "The assumptions are consistent"))))

;;; Returns a justification for false killing the assumption set.  This
;;; finds only one, and for the moment, doesn't necessarily return the
;;; informant.
;;; This should now call find-nogood-for-assumptions.  Who still calls why-nogood-assumptions?
;;; *****
(defun why-nogood-assumptions (assumptions &optional dont-lookup &aux env v)
  (setq env (if dont-lookup *contra-env* (find-or-make-env assumptions)))
;  (unless (env-contradictory env) (error "Assumptions are consistent!?"))
  (cond ((not (eq env *contra-env*)) (why-nogood-env env))
	((not (compatible-assumptions? assumptions (setq v (make-env-vector* assumptions))))
	 (do ((as assumptions (cdr as)))
	     ((null as) (error "Can't happen"))
	   (dolist (a (cdr as))
	     (when (vector-member (car as) (assumption-binary-vector a))
	       (return-from why-nogood-assumptions
		 (list '<UNKNOWN-INFORMANT> (car as) a))))))
	((setq env (new-contradictory-env-assumptions? assumptions (length assumptions)))
	 (why-nogood-env env))
;	(t (error "Uh --- I can't find a reason why the assumptions are inconsistent"))
	))

;;; Checks all known nodes to make sure they their label is right.
(defun check-node-labels ()
  (dolist (node *nodes*) (check-node-label node)))

;;; Inefficient but straight forward.  This checks only for missing envs right now.
(defun check-node-label (node &aux old-envs envs)
  (setq old-envs (append (node-envs node) nil))	;**** want non-cdr coded copy.
  (dolist (j (node-justifications node))
    (setq envs (user-general-weave nil *empty-env-list* (cdr j)))
    (dolist (env envs)
      (cond ((memq env old-envs))
	    ((dolist (old-env old-envs) (if (subset-env? old-env env) (return T))))
	    (t (error "Missing env ~A" env))))))

;;; Checks all known nodes have no nils in their label.

(defun check-nodes () (dolist (node *nodes*) (check-node-envs node)))

(defun check-node-envs (node)
  (if (memq nil (node-envs node)) (error "Node ~A has nil in its label" node)))

(defun check-env-table ()
  (dotimes (i (1+ *max-env-count*))
    (dolist (e (aref *environments* i))
      (check-an-env e))))

(defun check-contra-table () (walk-contradictions #'check-an-env))
      
(defun walk-contradictions (function &optional (>i 0))
  (dotimes (i (1+ *max-contra-count*))
    (unless (<= i >i)
      (walk-tree function (aref *nogood-trees* i)))))

;;; This misses binary nogoods.
(defun check-an-env (e &aux contra assumptions)
  (setq assumptions (env-assumptions e))
  (do ((assumptions assumptions next)
       (next (cdr assumptions) (cdr next)))
      ((null next))
    (unless (> (assumption-unique (car assumptions)) (assumption-unique (cadr assumptions)))
      (format T "~% ~A has its assumptions in bad order" e)))
  (setq contra (new-contradictory-env-assumptions? (env-assumptions e)
						   (env-count e)))
  (cond ((null contra)
	 (if (and (env-contradictory e) (eq (car (env-contradictory e)) 'SUBSUMED))
	     (format T "~% ~A is marked contradictory when it shouldn't be" e)))
	((null (env-contradictory e))
	 (format T "~% ~A is not marked contradictory when it contains the nogood ~A"
		 e contra)
	 (print e)
	 (print contra)
	 )
	((subsumed-nogood? e)
	 (if (eq e contra)
	     (format T "~% ~A is marked subsumed, but isn't" e)))
	((eq e contra))
	(t (format T "~% ~A is subsumed by ~A but not marked as such" e contra))))

;;; This makes sure a label is good.
(defun check-envs (envs)
  (do ((envs envs (cdr envs)))
      ((null envs))
    (do ((tail (cdr envs) (cdr tail)))
	((null tail))
      (if (memq (compare-env (car envs) (car tail)) '(SUBSET12 SUBSET21 EQUAL))
	  (error "Lossage")))))

;;; This checks for DOWNING's bug which has a screwed up cache on singleton envs.  This
;;; is inefficient but works.
(defun check-doubles (&aux env cache)
  (dolist (assumption *assumptions*)
    (setq env (assumption-env assumption)
	  cache (cdr (env-cons-env-cache env)))
    (if cache
	(maphash #'(lambda (other-assumption double-env)
		     (unless (and (memq other-assumption (env-assumptions double-env))
				  (memq assumption (env-assumptions double-env))
				  (> (assumption-unique other-assumption)
				     (assumption-unique assumption)))
		       (error "Screwed up cache")))
		 cache))))

;;; 
(defun check-assumptions ()
  (do ((as *assumptions* (cdr as)))
      ((null as))
    (dolist (e (assumption-envs (car as)))
      (cond ((> (env-count e) 1))
	    ((eq (car (env-assumptions e)) (car as)))
	    ((memq (assumption-env (car as)) (assumption-envs (car (env-assumptions e))))
	     (format T "~% Assumptions ~A and ~A are equivalent"
		     (car as) (car (env-assumptions e))))
	    (t (format T "~% Assumption ~A depends on assumption ~B"
		       (car as) (car (env-assumptions e))))))))

(defun ensure-monitoring ()
  #+Symbolics
  (unless meter:*pc-monitor-array*
    (format T "~% PC array does not exist---I'm creating one")
    (meter:make-pc-array (* sys:page-size 64.))
    (meter:monitor-all-functions)))

(defun nogood-distribution (&aux (total 0) count result)
  (dotimes (i (1+ *max-contra-count*))
    (setq count (count-tree (aref *nogood-trees* i)))
    (unless (= count 0)
      (incf total count)
      (push (cons i count) result)))
  (values result total))




;;;For ADB.
(defun nogoods-between (lower upper &aux result)
  (if (>= upper *max-contra-count*) (setq upper *max-contra-count*))
  (unless (<= upper lower)
     (do ((i lower (1+ i)))
	 ((> i upper))
       (setq result (append (collect-minimal-nogoods i) result))))
  result)

;;; Returns all minimal nogoods of size i.
(defun collect-minimal-nogoods (i) (collect-tree (aref *nogood-trees* i)))

;;; Acceptance tests of a new ATMS.  We should add a lot more here after time.

;;; bits/break *c012*:T/NIL *k*=2,3,4.

(defun accept ()
  (setq *c012* nil) (install-bits)
  (basic-accept)
  (install-break)
  (basic-accept)
  (setq *c012* T)
  (basic-accept)
  (install-bits)
  (basic-accept))

(defun basic-accept ()
  (setq *h4* nil *k* nil)
  (unless (= (length (freuder-1)) 2)
    (error "(freuder-1) generated wrong number of solutions"))
  (setq *h4* 1 *k* nil)
  (unless (= (length (freuder-1)) 2)
    (error "(freuder-1) generated wrong number of solutions"))
  (unless (= (length (freuder-2)) 6)
    (error "(freuder-2) generated wrong number of solutions"))
  (setq *h4* nil *k* nil)
  (unless (null (freuder-3))
    (error "(freuder-3) generated wrong number of solutions"))
  (setq *h4* 1 *k* 3)
  (unless (null (freuder-3))
    (error "(freuder-3) generated wrong number of solutions"))
  (setq *h4* nil *k* nil)
  (basic-basic-accept)
  (setq *h4* 1 *k* 2)
  (basic-basic-accept)
  (setq *k* 3)
  (basic-basic-accept)
  (setq *k* nil)
  (basic-basic-accept))

(defun basic-basic-accept ()

  (unless (= (length (n-queens 8)) 92.)
    (error "(n-queens 8) generated wrong number of solutions"))

  (unless (= (length (ib-1)) 1176.)
    (error "(ib-1) generated wrong number of solutions"))

  (unless (= (length (tn-queens 7)) 28.)
    (error "(tn-queens 7) generated wrong number of solutions"))

  (init-tms) (parity 9)
  
  (replay-all-trace-files))

(defun replay-all-trace-files ()
  (dolist (file '(">atms>trace>new-bug.trace"
		  ">atms>trace>ptest3-ir.trace"
		  ">atms>trace>ptest3-level.trace"
		  ">atms>trace>qpe-bug.txt"
		  ">atms>trace>trace.lisp"
		  ">atms>trace>v2-tms-tracea.txt"
		  ">atms>trace>v2-tms-traceb.txt"
		  ">atms>trace>v2-tms-tracec.txt"
		  ">atms>trace>gordon.lisp"
		  ))
    (replay file nil)))

;;; This runs short ATMS tests.
(defun short-test ()
  (dolist (file '(">atms>trace>ex1.trace"
		  ">atms>trace>ex2.trace"
		  ">atms>trace>ex3.trace"
		  ">atms>trace>ex4.trace"		  
		  ">atms>trace>ex5.trace"))
    (format T "~% Replaying file ~A" file)
    (replay file nil nil t)
    (profile-replay))
  (n-queens 8)
  (n-queens 9)
  (n-queens 10))

(defun benchmark ()
  (benchmark-files 3 '(">atms>trace>ex1.trace"
		       ">atms>trace>ex2.trace"
		       ">atms>trace>ex4.trace"
		       ">atms>trace>ex5.trace"))
  (benchmark-files 1 '(">atms>trace>ex3.trace"
		       ">atms>trace>ex7.trace"))
  (bench-queens)

  )

#+(AND Symbolics (not Symbolics-Release-6))
(defun benchmark-files (repeats files &aux (output zl:terminal-io))
  (let ((*standard-output* #'si:null-stream))
    (dolist (file files)
      (replay file nil nil T)
      (format output "~% ~D runs of ~A with gc-off but preemption allowed" repeats file)
      (si:gc-off)
      (dotimes (i repeats) (time (profile-replay nil nil)))
      (format output "~% ~D runs of ~A with gc-off preemption disallowed" repeats file)
      (dotimes (i repeats)
	(time (process:without-preemption (profile-replay nil nil))))
      (si:gc-on))))

#+(AND Symbolics (not Symbolics-Release-6))
(defun bench-queens ()
  (let ((*standard-output* #'si:null-stream))
    (si:gc-off)
    (dotimes (i 3) (time (n-queens 8)))
    (dotimes (i 3) (time (process:without-preemption (n-queens 8))))
    (dotimes (i 3) (time (dn-queens 8 t)))
    (dotimes (i 3) (time (process:without-preemption (n-queens 8 t))))))


(defun node (u)
  (dolist (n *nodes*) (if (eq (node-unique n) u) (return n))))

;;; So you can get a particular env.
(defun e (n)
  (catch 'E
    (dotimes (i (1+ *max-env-count*))
      (dolist (e (aref *environments* i)) (if (= (env-unique e) n) (throw 'E e)))
      (walk-contradictions #'(lambda (e) (if (= (env-unique e) n) (throw 'E e)))))))

(defun n (n)
  (catch 'E (dolist (node *nodes*) (if (= (node-unique node) n) (throw 'E node)))))

;;; Explanation facilities.

;;; Returns a list of justifications, one per node, all holding in env, which deduce node.
;;; Remember a justification here is: (consequence informant . antecedents)

;;; Slooooooww, but clean.  Finds one, not all, not the best explanation.  Thats much much
;;; harder. I haven't actually proved to myself that this works --- I need to sit
;;; in a corner and prove it to myself.  I think its clear that removing environment narrowing
;;; would make things work guaranteed.

;;; Make this a hash table if needed.
(defvar *node-justs* nil)			; (node informant node node ...)
						; (node) if not possible.
(defun explanation-for-node (node env &optional *node-justs*)
  (if (eq env *contra-env*) (error "Whadya mean bunkie?"))
  (cond ((env-contradictory env) (explanation-for-nogood-1 env))
	((explanation-for-node-1 node env nil))
	(t (error "No explanation found.")))
  (collect-explanation (if (env-contradictory env) env node) nil))

;;; This I suspect misses a few types of justifications for false for which the
;;; nogoods have to be consed.  So a little more work is required...
;;; This is pretty slow.
(defun explanation-for-nogood (nogood &aux base-env)
  (setq base-env (find-base-contradiction nogood))
  (explanation-for-node *contra-node* base-env))

(defun explanation-for-nogood-1 (nogood &aux reason result)
  (setq nogood (find-base-contradiction nogood))
  (cond ((setq result (assq nogood *node-justs*)) (cdr result))
	(t (setq reason (env-contradictory nogood))
	   (selectq (cadr reason)
	     (RESOLUTION (setq result (explain-resolution-result nogood reason))
			 (push `(,nogood RESOLUTION .,result) *node-justs*)
			 (dolist (nogood result) (explanation-for-nogood-1 nogood)))
	     (T (push (list nogood 'BASE *contra-node*) *node-justs*)
		(explanation-for-node-1 *contra-node* nogood nil))))))

;;; Reason looks like (BASE RESOLUTION OR (a . nogood-1) ...)
(defun explain-resolution-result (env reason &aux result table)
  env
  (setq table (or-nogood-explanations (third reason)))
  (dolist (pair (fourth reason)) (push (cadr (or (assoc pair table :test #'EQUAL)
						 (error "not found")))
				       result))
  result)

;  (explain-resolution-result-1 env (or-disjuncts (caddr reason)) (or-nogoods (caddr reason))))

(defun explain-resolution-result-1 (env disjuncts nogoods &aux result)
  (if (null disjuncts)
      T
      (dolist (nogood-1 (cddr (assq (car disjuncts) nogoods)))
	(cond ((env-contradictory nogood-1))
	      ((null (setq result (explain-resolution-result-1 env (cdr disjuncts) nogoods))))
	      (t (return (cons (find-nogood-for-assumptions (cons (car disjuncts)
								  (env-assumptions nogood-1)))
			       (unless (eq result T) result))))))))

(defun collect-explanation (node explanation &aux j)
  (unless (assq node explanation)
    (setq j (assq node *node-justs*))
    (push j explanation)
    (dolist (n (cddr j)) (setq explanation (collect-explanation n explanation))))
  explanation)

;;; Results:
;;;   NIL = failed to find any explanation.
;;;   List = the list of justifications needed.

;;; Collected justifications is superfluous, just pull result out of *node-justs*.
(defun explanation-for-node-1 (node env pending-nodes &aux result)
  (cond ((setq result (assq node *node-justs*)) (cdr result))
	((memq node pending-nodes) nil)
	((and (assumption? node) (memq node (env-assumptions env)))
	 (push (list node 'ASSUMING-MYSELF node) *node-justs*))
	(t (dolist (j (node-justifications node))
	     (setq result (explanation-for-just j env (cons node pending-nodes)))
	     (when result
	       (setq result (cons node j))
	       (push result *node-justs*)
	       (return-from EXPLANATION-FOR-NODE-1 result)))
	   ;; Signal there is no justification for this.
	   (push (ncons node) *node-justs*)
	   nil)))

(defun explanation-for-just (j env pending-nodes &aux aenv)
  (zf (env-contradictory env)
      (dolist (n (cdr j))
	(unless
	  (zf (setq aenv (true-in? n env))
	      (or (explanation-for-node-1 n aenv pending-nodes)
		  (error "Impossible"))
	      (explanation-for-node-1 n env pending-nodes))
	  (return-from EXPLANATION-FOR-JUST nil)))
      (dolist (n (cdr j)) (zf (i-out? n) (return-from EXPLANATION-FOR-JUST nil)))
      ;; Be very carefull, circularities can trip us up here all over.
      (unless (dolist (aenv (antecedent-envs j))
		(and (subset-env? aenv env)
		     (null (dolist (n (cdr j))
			     (unless (explanation-for-node-1 n aenv pending-nodes)
			       (return T))))
		     (return T)))
	(return-from EXPLANATION-FOR-JUST nil)))
  T)

;;; This skips the singleton and binary nogoods which isn't right.
(defun explain-all-nogoods ()
  (walk-contradictions #'(lambda (nogood)
			   (unless (explanation-for-nogood nogood)
			     (error "Can't explain the contradiction: ~A"nogood)))))

;;; Tests for union-envs.

(defun test-u (l) (dolist (p l) (union-envss (car p) (cadr p))))

(defun test-us (l &aux r1 r2)
  (dolist (p l) (setq r1 (union-envss1 (car p) (cadr p))
		      r2 (union-envss2 (car p) (cadr p)))
	  (unless (same-setp r1 r2) (error "Bug in some union-envs"))))

(defun same-setp (s1 s2)
  (cond ((equal s1 s2))
	((not= (length s1) (length s2)) nil)
	(t (dolist (e1 s1 T) (unless (memq e1 s2) (return-from SAME-SETP nil))))))

