;;; -*- Syntax:Common-Lisp; Mode: LISP; Package: (TMS Lisp 1000.); Base: 10. -*-

"(c) Copyright 1986, 1987, 1988, 1989 Xerox Corporation.  All rights reserved.  Subject to
the following conditions, permission is granted to use and copy this
software and to prepare derivative works:  Such use, copying or
preparation of derivative works must be for non-commercial research or
educational purposes; each copy or derivative work must include this
copyright notice in full; a copy of each completed derivative work must
be returned to:  DEKLEER@XEROX.COM (Arpanet) or Johan de Kleer,
Xerox PARC, 3333 Coyote Hill Road, Palo Alto, CA 94304.  This software
is made available AS IS, and Xerox Corporation makes no warranty about
the software or its performance."

(in-package 'tms)

;;; This file contains all the functions used to do label propagation.

;;; For many examples it pays incredibly to find to size 1 and size 2 contradictions first.
;;; In fact, more general the right order is for each increaseing n find the contradictions
;;; of size n, then the environments of size n, etc.  But I suspect that is all simply 
;;; to much book-keeping to be worth it.
(defun update-node1 (node new-envs complete-envs 
		     &aux start-time start-count contra-count
		     start-contra-counter)
  (when *update-trace*
    (setq start-time (get-internal-run-time)
	  start-count *env-counter*
	  contra-count 0
	  start-contra-counter *contra-counter*))
  (cond (*c012*
	 (update-node-c012 node new-envs nil)
	 (setq complete-envs (sanitize-envs complete-envs))
	 (if (setq new-envs (sanitize-envs new-envs))
	     (update-node node new-envs complete-envs)))
	(t (update-node node new-envs complete-envs)))

  (when *update-trace*
    (format T "~% Update time is:~D seconds" (time-taken start-time))
    (format T "~% ~D new minimal nogoods" (- *contra-counter* start-contra-counter))
    (if *going-nodes* (format T "~% ~D going nodes" (length *going-nodes*)))
    (format T "~% ~D new envs created" (- *env-counter* start-count))
    (do ((i (1+ start-count) (1+ i)))
	((> i *env-counter*))
      (if (e i) (incf contra-count)))
    (format T "~% Of which ~D are consistent." contra-count)))



;;; This needs to incrementally sanitize****Like above.********
(defun propagate-clauses (node new-envs
			  &aux positives negatives flag open)
  (dolist (clause (n-a-nclauses node))
    ;; Make sure the clause is unit-open somewhere --- a big assumption.*****
    (setq negatives (first clause)
	  positives (second clause))
    ;; Check if the clause is active---perhaps this short be part of the clause
    ;; data structure.******

    ;; Test this check later.

    (setq open nil)
          
    ;; If clause is satisfied, or has more than one open literal it is useless.
    (cond ;; Check whether the clause is useless right now.
	  ((or (dolist (n negatives)
		 (cond ((in? n))
		       ((i-false? n) (return T))
		       (open (return T))
		       (t (setq open n))))
	       (dolist (p positives)
		 (cond ((i-true? p) (return T))
		       ;;???
		       ))))
	  
	  ((cdr positives))
	  ;;**how does open affect this?
	  (positives (update-node-envs (car positives) negatives node new-envs))
	  ;; If there is only one open, and everything is negative, its easy:
	  (open (update-node-envs (n-a-neg open) (remove open negatives) node new-envs))
	  (t (setq flag nil)
	     (dolist (n negatives)
	       (unless (or (eq n node) (i-true? n))
		 (if (n-a-neg n)
		     (update-node-envs (n-a-neg n) (remove n negatives) node new-envs)
		     (error "~A has no negation" n))))
;		     (setq flag T)
	     (if flag (update-node-envs *contra-node* negatives node new-envs))))))

;;; A consequent looks like (consequent-node . justification)
;;; New environments have been discovered for the antecedent node.
;;; *** why does this look at justifications only once. 

;;; This returns the possibly side-effected new-envs.
(defun update-node-envs (node antecedent-nodes antecedent-node new-envs
			 &aux justifications)
  (setq justifications (node-justifications node))
  (cond ((node-contradictory node)
	 (unless (cdr antecedent-nodes)
	   (error "Contradiction should have been detected earlier"))
	 (weave-for-false new-envs antecedent-nodes antecedent-node)
	 new-envs)
	((i-true? node) new-envs)
	
	;Used to do this in above condition.
	;(contradictory-envs new-envs (list 'BASE 'LABEL-FOR-FALSE consequent)))
	
	;; If there are more than one antecedent nodes in this justification, just
        ;; recompute everything.  (Should be optimized more****.)
	((cdr antecedent-nodes)
	 (if *trace* (format T "~% Updating envs of ~A because of new antecedent envs ~A"
			     (node-string node) new-envs))
	 ;; This happens so often that its worth a pre-check.
	 (dolist (n antecedent-nodes)
	   (if (i-out? n) (return-from UPDATE-NODE-ENVS new-envs)))
	 
	 ;; This depends crucially on general-weave returning FRESH list structure
	 ;; **always**.  Because new-envs gets side-effected in this algorithm.
	 (multiple-value-bind (known-envs new-envs)
	     (general-weave (node-envs node) new-envs antecedent-nodes antecedent-node)
;	   (if new-envs (format T "+") (format T "-"))
	   (when new-envs
	     (update-node node new-envs known-envs)))
	 new-envs)
	;; If there is more than one justification, the envs of the consequent could
	;; already be updated.  Also, the new-envs could eliminate some environments
	;; of the label.  If the node is a member of some classes or has a negation
	;; its label can be affected by other mechanisms.
	((or (cdr justifications) (node-classes node) (n-a-neg node) (node-pclauses node))
	 (update-node-envs-simple node new-envs)
	 new-envs)
	;; If there is only one justification, and it has only one antecedent node, just
        ;; copy the update forward with doing any work.  But remember each node had
	;; better have a unique list structure for its label.  This is save ONLY if this
	;; this justification is the **only** way this node can be affected.
	(t (update-node node new-envs (fcopylist (node-envs antecedent-node))))))

;;; And new-envs to node, do all subsumptions.  This duplicates some of the work
;;; of change-node-envs.  Is this efficiency hack really worth it?
(defun update-node-envs-simple (node new-envs &aux node-envs new-new-envs)
  ;; This fcopylist is necessary, because envs gets rplaca'd.
  (setq node-envs (fcopylist (node-envs node)))
  (dolist (new-env new-envs)
    (unless (or (env-contradictory new-env)
		(do ((known-envs node-envs (cdr known-envs)))
		    ((null known-envs) nil)
		  (if (car known-envs)
		      (selectq (compare-env (car known-envs) new-env)
			(SUBSET21 (rplaca known-envs nil))
			((EQUAL SUBSET12) (return T))))))
      (push new-env new-new-envs)))
  (when new-new-envs
    (setq node-envs (fdelqa nil node-envs))
;    (setf (node-envs node) node-envs)
    ;; Really should do an fcopylist here to get a cdr-coded list.
    (update-node node new-new-envs (append new-new-envs node-envs))))

;;; Could be a macro if slow.
;;; Note that this should never be called if the node is an
;;; ****** the full test on assumption-nogoods hasn't perculated to c012.
;;; This assumes complete-envs is fresh data structure.
;;; Note this is the only place change-node-envs is called so change-node-envs
;;; should rely on new-envs too.
(defun update-node (node new-envs complete-envs &aux not-first consequent contra-counter neg)
  ;; A primitive focussing mechanism, could be much more efficient.  Could recycle
  ;; conses ****.
  (when *foci*
    (multiple-value-setq (new-envs complete-envs)
      (update-node-filter node new-envs complete-envs))
    (unless new-envs (return-from UPDATE-NODE nil)))
  (when (assumption? node)
    (multiple-value-setq (new-envs complete-envs)
      (update-assumption node new-envs complete-envs))
    (unless new-envs (return-from UPDATE-NODE nil)))
  (change-node-envs node complete-envs)
;;; Remember we are running depth first, so things can change right under us.  So
;;; for every consequent check that the new-envs set is really still new and not
;;; contradictory.  Note that update-node-consequents will clobber new-envs.
;;; New-envs better be new, or else this will infinite loop.  This algorithm will
;;; side-effectively remove any inconsistent environments on new-envs.  It returns
;;; the changed new-envs also.

;;; This code is based on the hypothesis that unless resolution happens underneath,
;;; a node-envs can never change underneath this algorithm.  This is clearly false
;;; if *h45* is on and resolutions are happening underneath.  The theorem holds because
;;; by necessity new-envs in recursive calls to update-node-consequents must grow
;;; monitonically.  It might be worth maintaining a stack of nodes whose environment
;;; we are updating and simply ignore circular updates.

  (if (node-contradictory node) (return-from UPDATE-NODE new-envs))
  ;; In the new regime, justifications aren't constructed for negations, this code
  ;; ensures that negations are properly processed.
  (cond ((null (setq neg (n-a-neg node))))
	((null (n-a-envs neg)))
	(T (setq not-first T contra-counter *contra-counter*)
	   (dolist (e new-envs)
	     (unions-for-false e (n-a-envs neg) '(NEGATION)))))
  ;; In the new regime, justifications aren't contructed for unique-value classes.
  ;; ***** This only assumes an exclusion class.
  (dolist (class (node-classes node))
    (when (class-exclusive class)
      (dolist (other (class-nodes class))
	(unless (or (eq other node) (i-false? other))
	  (unless not-first (setq not-first T contra-counter *contra-counter*))
	  (cond ((n-a-neg other)
		 (update-node-envs-simple (n-a-neg other) new-envs))
		(t (dolist (e new-envs)
		     (unions-for-false e (n-a-envs other)
				       `(EXCLUSION ,node ,other)))))))))
  ;; Look at the clauses in which it appears with opposite sign.
  (when (n-a-nclauses node)
    ;; Forget this optimization.
    (setq not-first T contra-counter *contra-counter*)
    (propagate-clauses node new-envs))
  ;; Process conventional justifications.
  (when (setq consequent (node-consequents node))
    (if *trace* (format T "~% Adjusting consequents of node ~S because of new found envs: ~A"
			node new-envs))
    (dolist (c consequent)
      (unless (i-true? (car c))
	(cond (not-first
	       ;; First check to see whether any of the new-envs became inconsistent.
	       ;; But only make this check if we have to.  
	       (unless (= contra-counter *contra-counter*)
		 (do ((new-env new-envs (cdr new-env)))
		     ((null new-env))
		   (when (env-contradictory (car new-env))
		     (rplaca new-env nil)))
		 (unless (setq new-envs (fdelqa nil new-envs))
		   (return-from UPDATE-NODE nil)))
	       (if *h45*
		   (setq new-envs (or (cleanup-for-h45 node new-envs)
				      (return-from UPDATE-NODE nil)))))
	      (t (setq not-first T)))
	;; *contra-counter* will increment if any envs become marked nogood.
	(setq contra-counter *contra-counter*)
	(unless (setq new-envs
		      (update-node-envs (car c) (cddr c) node new-envs))
	  (return-from UPDATE-NODE nil))))
    (if *trace* (format T "~% Finishing consequents of node ~A" (node-string node))))
  new-envs
  )

;;; Keep delayed minimal???? ***** is it worth it?  Probably not.  But think
;;; about it some day.
(defun update-node-filter (node new-envs complete-envs &aux update change delayed odelayed)
  (setq delayed (node-blocked node)
	odelayed delayed)
  (do ((new new-envs (cdr new)))
      ((null new))
    (unless (dolist (f *foci*)
	      (when (subset-env? (car new) f)
		(setq update T)
		(return T)))
      (setq change T)
      (push (car new) delayed)			;****** cons could be recycled.
      (rplaca (memq (car new) complete-envs) nil)
      (rplaca new nil)))
  (when change
    (setf (node-blocked node) delayed)
    (unless odelayed (push node *foci-blocked*)))
  (unless update (return-from UPDATE-NODE-FILTER nil))
  (unless change (return-from UPDATE-NODE-FILTER (values new-envs complete-envs)))
  (values (fdelqa nil new-envs) (fdelqa nil complete-envs)))

;;; Tell the ATMS that foci changed.
;;; This can be made much more efficient if necessary.
;;; This **** does not save conses.
;;; Supersets of old foci are not new foci.
(defun change-foci (new-foci &aux really-new-foci new-envs known-envs blocked start-time
		                  *going-nodes*)
  (setq start-time (get-internal-run-time))
  (dolist (f new-foci) (unless (memq f *foci*) (push f really-new-foci)))
  (setq *foci* new-foci)
  (when really-new-foci
    (do ((n *foci-blocked* (cdr n)))
	((null n))
      (setq new-envs nil
	    known-envs (node-envs (car n))
	    blocked (node-blocked (car n)))
      (do ((e blocked (cdr e)))
	  ((null e))
	(when (and (car e)
		   (dolist (f really-new-foci) (if (subset-env? (car e) f) (return T))))
	  (cond ((env-contradictory (car e)))
		((do ((known-envs known-envs (cdr known-envs)))
		     ((null known-envs) nil)
		   (if (car known-envs)
		       (selectq (compare-env (car known-envs) (car e))
			 ;; Flag --- we need only check for SUBSET21 if we find one.
			 ;; so this is somewhat inefficient.
			 ;; Also, we need to do memq check until up old known-envs
			 (SUBSET21 (setq known-envs (fdelq1 (car known-envs) new-envs))
				   (rplaca known-envs nil))
			 ((EQUAL SUBSET12) (return T))))))
		(t (push (car e) known-envs)
		   (push (car e) new-envs)))
	  (rplaca e nil)))
      (setf (node-blocked (car n)) (fdelq1 nil blocked))
      (when new-envs (update-node (car n) new-envs (fdelqa nil known-envs))))
    (do ((n *foci-blocked* (cdr n)))
	((null n))
      (unless (node-blocked (car n)) (rplaca n nil)))
    (setq *foci-blocked* (fdelqa nil *foci-blocked*)))
  (if *going-nodes* (process-changed-nodes))
  (format T "~% Refocus time is:~D seconds"
	  (time-taken start-time)))

;;; In general we can do far smart things.  This is a special-case optimization.
;;; For binary nogoods.  This can be made more efficient.  The fdelqa here is stupide.
(defun update-assumption (node new-envs complete-envs &aux one-bad uncons)
  (do ((env new-envs (cdr env)))
      ((null env))
    (cond ((vector-intersection? (assumption-binary-vector node)
				 (env-vector (car env)))
	   (contradictory-env (car env) '(BASE ASSUMPTION-LABELING))
	   (setq one-bad T)
	   (rplaca env nil))))
  (when one-bad
    (setq new-envs (sanitize-envs new-envs))
    (unless new-envs (return-from UPDATE-ASSUMPTION nil))
    (setq complete-envs (sanitize-envs complete-envs)
	  one-bad nil))
  (dolist (nogood (assumption-nogoods node))
    (setq uncons nil)
    (cond ((assumption? nogood))
	  ((subsumed-nogood? nogood))
	  (t (dolist (env new-envs)
	       (cond ((and (null env) (env-contradictory env)))
		     (t (unless uncons (setq uncons (uncons-env nogood node)))
			;; Actually one-bad needs to be set only if we find a new nogood.
			(setq one-bad T)
			(union-for-false uncons env 'NEEDS-WORK)))))))
  (when one-bad
    (setq new-envs (sanitize-envs new-envs))
    (unless new-envs (return-from UPDATE-ASSUMPTION nil)))
  (when one-bad (setq complete-envs (sanitize-envs complete-envs)))
  (if *h45*
      (setq new-envs (or (cleanup-for-h45 node new-envs) (return-from UPDATE-ASSUMPTION nil))))
  (values new-envs complete-envs))


;;; This should be called after every call to contradictory-env on the current env
;;; set as that can cause arbitrary envs hacking.  *h45* is one such example.  
;;; Is resolution another???? Check.  Returns the updated new-envs.  Can be grossly
;;; optimized****

	     ;; Now remove all new-envs which have become subsumed for this node.
	     ;; **** this is totally stupid.  All you really have to check is to make
	     ;; sure each new-env is still on node-envs.  Thats absolutely all!!!!!!.
	     ;; Thus, for all cases this is inefficient code***************************
	     ;; And if h45 happens above, like process-queued-nogoods.  This can 
	     ;; never happen.

;;; This used to be a subset-env?, but thats not good enough.
(defun cleanup-for-h45 (node new-envs &aux node-envs)
  (setq node-envs (node-envs node))
  (do ((new-env new-envs (cdr new-env)))
      ((null new-env))
    (if (car new-env)
	(dolist (node-env node-envs)
	  (when (subset-env? node-env (car new-env))
	    (rplaca new-env nil)
	    (return)))))
  (setq new-envs (fdelqa nil new-envs)))

;;; This search is specifically organized to find contradictions of size 1 or size 2 first.
;;; It may fortuitously find other contradictions (may not be worth it?)
;;; but it will never actually
;;; create a contradictory environment, because that has shown in practice to be
;;; too expensive.

;;; Returns NIL if everything in new-envs has become contradictory or is to big.
;;;; ***** incompletely implemented: Returns T more often than necessary.
(defun update-node-c012 (node new-envs stack &aux continue)
  (if (memq node stack) (return-from UPDATE-NODE-C012 nil))
  ;; In general we can do smart things.  This is a special-case optimization.
  ;; For binary nogoods.  This can be made more efficient.  The fdelqa here is stupid.
  ;; *** this check may be thus removable from the normal update-node***
  (when (assumption? node)
    (do ((env new-envs (cdr env)))
	((null env))
      (cond ((null (car env)))
	    ((env-contradictory (car env)))
	    ((vector-intersection? (assumption-binary-vector node)
				   (env-vector (car env)))
	     (contradictory-env (car env) '(BASE ASSUMPTION-LABELING)))
	    ((< (env-count (car env)) 3.) (setq continue T))))
    (unless continue (return-from UPDATE-NODE-c012 nil)))
  (cond ((node-contradictory node)
	 (dolist (env new-envs)
	   (and env
		(not (env-contradictory env))
		(contradictory-env env '(BASE UPDATE-NODE-C012))))
	 T)
	(t (update-node-consequents-c012 node new-envs (cons node stack)) ;*** stack cons??
	   T)))

(defun update-node-consequents-c012 (node new-envs stack &aux consequent)
  (unless (null (setq consequent (node-consequents node)))
    (do () (nil)
      (unless (update-node-envs-c012 (car consequent) node new-envs stack)
	(return-from UPDATE-NODE-CONSEQUENTS-C012 nil))
      (unless (setq consequent (cdr consequent))
	(return-from UPDATE-NODE-CONSEQUENTS-C012 nil)))))

;;; This is only looking for nogoods of size 0, 1, 2 and those existing
;;; environments it fortuitously stumbles across.
;;; *** should it update labels???????????????***** this version does not!!!
;;; *** maybe it should update labels which don't cause cres??

;;; Returns NIL if new-envs is totally contradictory.  Not implemented enough.

(defun update-node-envs-c012 (consequent antecedent-node new-envs stack
			 &aux justifications node antecedent-nodes)
  (setq node (first consequent) antecedent-nodes (cddr consequent))
  (setq justifications (node-justifications node))
  (cond ((node-contradictory node)
	 (unless (cdr antecedent-nodes)
	   (error "Contradiction should have been detected earlier"))
	 (weave-for-false-c012 new-envs (cdr consequent) antecedent-node)
	 T)
	((i-true? node))
	((cdr antecedent-nodes)
	 ;; This is probably faster than weaving.  Dunno.  The idea is we
	 ;; don't want to cons up all these stupid lists all the time.
	 ;; ** assumptions
	 (dolist (env new-envs)
	   (avoid-weaving consequent env antecedent-nodes antecedent-node stack))
	 T)
	(t (update-node-c012 node new-envs stack))))

(defun avoid-weaving (consequent env nodes antecedent-node stack &aux env1)
  ;; Punt if we can't do anything interesting.
  (if (or (env-contradictory env) (> (env-count env) 2)) (return-from AVOID-WEAVING nil))
  (prog ()
	start
	;; Skip to the first node which can tell us anything.
	(do nil ((null nodes)
		 ;; This test can be applied every union***********
		 ;; Can be highly optimized.***********************
		 ;; we can check This real easy.************
		 (dolist (known-env (node-envs (car consequent)))
		   (and (< (env-count known-env) 3)
			(subset-env? known-env env)
			(return-from AVOID-WEAVING nil)))
		 (update-node-c012 (car consequent) (list env) stack)
		 (return-from AVOID-WEAVING nil))
	  (cond ((i-true? (car nodes)))
		((eq (car nodes) antecedent-node) (setq antecedent-node nil))
		(t (return nil)))
	  (setq nodes (cdr nodes)))
	(when (eq env *empty-env*)
	  (setq env1 nil)
	  (dolist (env (node-envs (car nodes)))
	    (cond ((> (env-count env) 2))
		  ((null env1) (setq env1 env))
		  (T (avoid-weaving consequent env (cdr nodes) antecedent-node stack))))
	  (cond (env1 (setq env env1 nodes (cdr nodes)) (go start))
		(t (return nil))))
	;; *** we could have a version of union which took three arguments!.
	(setq env1 nil)
	(dolist (new-env (node-envs (car nodes)))
	  (cond ((> (env-count new-env) 2))
		;;*** could be optimized because are 1 or 2.
		((env-contradictory (setq new-env (2-union-env env new-env))))
		((null env1) (setq env1 new-env))
		(T (avoid-weaving consequent new-env (cdr nodes) antecedent-node stack))))
	(when env1 (setq env env1 nodes (cdr nodes)) (go start))))

;;; Env is new for antecedent-node, compute a new env(s) for consequent-node.
;;; I tries to avoid calling general-weave-c012 at all.
;;;****** is the first argument to update-node-c012 always redundant?***
;;;****** Make second argument always be a singleton.
;;;****** incorporate ideas of weaving-for-false, weaving-for-false-c012 here.
;;;****** this is a bit of a crock.  Weaving-for-false is the best example.
(defvar *justification*)

;;; This only attempts to find contradictions of size 0, 1, 2.
(defun weave-for-false-c012 (envs *justification* node-to-ignore)
  (cond ;; If any antecedents are out, punt.
	((dolist (n-a (cdr *justification*))
	   (cond ((eq n-a node-to-ignore))
		 ((i-out? n-a) (return T)))))
	(node-to-ignore
	 (dolist (e envs)
	   (weaving-for-false-c012 (cdr *justification*) e node-to-ignore)))
	(t (weaving-for-false-c012 (cdr *justification*) *empty-env* nil))))

;;;****** The 2-union-env could be fixed here.  We could do this depth first after all.
;;; consing up the doubles is easy.
(defun weaving-for-false-c012 (nodes env node-to-ignore &aux envs)
  (prog ()
    loop
       (cond ((env-contradictory env))
	     ((null nodes)
	      (contradictory-env env
				 (cons 'BASE (cons 'WEAVING-FOR-FALSE *justification*))))
	     ((or (null (cdr nodes))
		  (and (null (cddr nodes)) (eq (cadr nodes) node-to-ignore)))
	      (if (unions-for-false-c012 env (node-envs (car nodes)) *justification*)
		  (return nil)))
	     ((eq (car nodes) node-to-ignore)
	      (setq nodes (cdr nodes) node-to-ignore nil)
	      (go loop))
	     ((cdr (setq envs (node-envs (car nodes))))
	      (dolist (e envs)
		(weaving-for-false-c012 (cdr nodes) (2-union-env env e) node-to-ignore)
		(if (env-contradictory env) (return nil))))
	     (t (setq env (2-union-env env (car envs))
		      nodes (cdr nodes))
		(go loop)))))

(defun unions-for-false-c012 (env envs reason)
  (cond ((env-contradictory env))
	((memq env envs) (contradictory-env env `(BASE UNIONS-FOR-FALSE-C012 ,reason)) T)
	((eq *empty-env* env)
	 (print "Should be optimized out earlier")
	 (dolist (e envs) (contradictory-env e `(BASE UNIONS-FOR-FALSE-C012 ,reason)))
	 nil)
	((eq (car envs) *empty-env*)
	 (print "Should be optimized out")
	 (contradictory-env env `(BASE UNIONS-FOR-FALSE-C012 ,reason))
	 T)
	((= (env-count env) 1)
	 (dolist (e envs)
	   (cond ((null e))
		 ((env-contradictory e))
		 ((= (env-count e) 1)
		  (double-for-false
		    (car (env-assumptions e)) (car (env-assumptions env)) reason)
		  (if (env-contradictory env) (return T)))
		 ((vector-member (car (env-assumptions env)) (env-vector e))
		  (contradictory-env env `(BASE UNIONS-FOR-FALSE-C012 ,reason))
		  (if (env-contradictory env) (return T))))))
	(t (dolist (e envs)
	     (cond ((null e))
		   ((env-contradictory e))
		   ((= (env-count e) 1)
		    (when (vector-member (car (env-assumptions e)) (env-vector env))
		      (contradictory-env env `(BASE UNIONS-FOR-FALSE-C012 ,reason))
		      (return T)))
		   ((vector-subset (env-vector env) (env-vector e))
		    (contradictory-env env `(BASE UNIONS-FOR-FALSE-C012 ,reason))))))))

;;; This weaver only bothers to return new 0, 1, 2 size environments.  
(defun general-weave-c012 (known-envs new-partial-envs consequent
			   &optional (node-to-ignore 'IGNORE) &aux new-new-partial-envs)
  (cond ((equal *empty-env-list* known-envs) nil)
	((dolist (n-a consequent)
	   (and (neq n-a node-to-ignore) (i-out? n-a) (return T)))
	 nil)
	(t (dolist (new-env new-partial-envs)
	     (cond ((null new-env))
		   ((env-contradictory new-env))
		   ((> (env-count new-env) 2.))
		   ((dolist (known known-envs)
		      (if (subset-env? known new-env) (return T))))
		   (t (push new-env new-new-partial-envs))))
	   (if new-new-partial-envs
	       (nweave-c012 consequent new-new-partial-envs node-to-ignore known-envs)))))

(defun nweave-c012 (nodes-to-process partial-envs node-to-ignore known-envs
		    &aux npartial-envs)
  ;; Try to handle the extremely simple common case where everything is a singleton,
  ;; and dispatch back to the general case if you lose.
  ;; *** we could have many difference versions of nweave-. for each case.
  ;; *** no need to return to the general general case.
  ;; *** if the general case reduces envs to one env, it could dispatch
  ;; *** to the easy case.
  (cond ((and (null known-envs) (null (cdr partial-envs)))
	 (nweave-simple-c012 nodes-to-process (car partial-envs) node-to-ignore))
	((equal partial-envs *empty-env-list*);;********** be much better.
	 (do ((nodes-to-process nodes-to-process (cdr nodes-to-process)))
	     ((null nodes-to-process) partial-envs)
	   (cond ((i-true? (car nodes-to-process)))
		 ((eq (car nodes-to-process) node-to-ignore))
		 (t (setq npartial-envs (fcopylist (n-a-envs (car nodes-to-process))))
		    (if npartial-envs
			(return (weave1-c012 (cdr nodes-to-process) npartial-envs
					node-to-ignore known-envs))
			(return nil))))))
	(t (weave1-c012 nodes-to-process partial-envs node-to-ignore known-envs))))

(defun nweave-simple-c012 (nodes-to-process env node-to-ignore
			   &aux node nenv env1 env2 envs comp last-nil oenvs)
  (do ((nodes-to-process nodes-to-process (cdr nodes-to-process)))
      ;; We could with crockery save this cons.
      ((null nodes-to-process) (list env))
    ;; If the node is to be ignored, ignore it.
    (unless (eq node-to-ignore (setq node (car nodes-to-process)))
      (setq nenv (car (n-a-envs node)))
      (cond ;; If the node is true, it contributes nothing.
	    ((eq nenv *empty-env*))
	    ;; If the next node has a simple label, its easy.
 	    ((null (cdr (node-envs node)))
	     (setq env (2-union-env env nenv))
	     (if (env-contradictory env) (return nil)))
	    ;; Here things get slightly dicey.  Lets hope that only one of the
	    ;; envs we have to construct survive.
	    (t (dolist (nenv (node-envs node))
		 (setq env1 (2-union-env env nenv))
		 (cond ;; If the result is contradictory we are in luck.
		       ((env-contradictory env1))
		       ;; If we haven't got any consistent envs yet, we are in luck.
		       ((null env2) (setq env2 env1))
		       ;; Things are not hopless yet, they might compare.
		       ((setq comp (compare-env env1 env2))
			(selectq comp
			  (EQUAL) ;;Damn Cl.
			  (SUBSET21)
			  (SUBSET12 (setq env2 env1))))
		       ;; The screw case, things are no longer simple.  We just have
		       ;; to byte the bullet and finish our work, and then trap
		       ;; out to the general case again.
		       ((null envs) (setq envs (list env1 env2)))
		       ;; We have to check for subsumption, in the usual way.
		       ;; Same old theorems apply.
		       (t (setq oenvs envs last-nil nil)
			  (do ((envs envs (cdr envs)))
			      ((null envs)
			       (if last-nil
				   (rplaca last-nil env1)
				   (push env1 oenvs)))
			    (cond ((null (car envs)) (setq last-nil envs))
				  (t (selectq (compare-env env1 (car envs))
				       ;; If the label is minimal, you could not have hacked
				       ;; envs at all yet.  This env does nothing.
				       (EQUAL (return))
				       (SUBSET12 (rplaca envs env1)
						 (do ((envs (cdr envs) (cdr envs)))
						     ((null envs))
						   ;; Can't find yourself if label was min.
						   (when (and (car envs)
						       (proper-subset-env? env1 (car envs)))
						     (rplaca envs nil)
						     (setq last-nil envs)))
						 (when last-nil
						   (fdelqa nil oenvs)
						   (unless (cdr oenvs)
						     (setq env2 (car oenvs) oenvs nil)
						   (return))))
				       (SUBSET21 (return))))))
			  (setq envs oenvs))))
	       ;; If envs is non-nil at this point, we have lost and can't use
	       ;; the simple approach any more: There is more than one env in the
	       ;; label so far.  Thus, we dispatch back to the general case.
	       (cond ((null envs) (if env2 (setq env env2 env2 nil) (return nil)))
		     ;; We don't have to do the general case if no nodes left.
		     ((or (null (cdr nodes-to-process))
			  (and (null (cddr nodes-to-process))
			       (eq (cadr nodes-to-process) node-to-ignore)))
		      (return envs))
		     (t (return
			  (nweave (cdr nodes-to-process) envs node-to-ignore nil)))))))))

;;; This can be incredibly optimized.  This is actually somewhat stupid.

(defun weave1-c012 (nodes-to-process partial-envs node-to-ignore known-envs
	       &aux new-env npartial-envs flag new)
  (dolist (node nodes-to-process)
    (setq npartial-envs nil)
    (cond ((eq node node-to-ignore))
	  ((i-true? node))
	  ;; If the label has exactly one environment, cross product is simpler.
	  ((null (cdr (node-envs node)))
	   (setq new nil)
	   ;; This loop updates the partial-envs list in place.  This works because the
	   ;; new partial-envs can't be any bigger than the original one.
	   ;;**** there will be so few, so why bother*** producing a copy?
	   (do ((penv partial-envs (cdr penv)))
	       ((null penv))
	     (cond ((null (car penv)))
		   ((or (env-contradictory (car penv)) (> (env-count (car penv)) 2.))
		    (rplaca penv nil))
		   ;; If the new union is inconsistent, delete it from the partial result.
		   ((env-contradictory
		      (setq new-env (2-union-env (car penv) (car (node-envs node)))))
		    (rplaca penv nil))
		   ;; If the new union is subsumed by what is already known, forget it.
		   ((dolist (known-env known-envs)
		      (if (subset-env? known-env new-env) (return T)))
		    (rplaca penv nil))
		   ;; Otherwise, scan the updated prefix of partial-envs.  When you
		   ;; find old self, replace it with your new self.
		   (t (setq flag nil)
		      (do ((penvs partial-envs (cdr penvs)))
			  ((eq penvs penv) (setq new T) (rplaca penvs new-env))
			(cond ((null (car penvs)))
			      ;; This is based on the theorem that if the prefix is
			      ;; always minimal, then if you have found an element prefix
			      ;; which you subsume, you cannot find yourself, nor any
			      ;; element which subsumes you.
			      (flag (if (subset-env? new-env (car penvs))
					(rplaca penvs nil)))
			      ;; Scan the prefix, to ensure that minimality is
			      ;; maintained.  If you ever find an element you subsume,
			      ;; then the much easier previous check is faster.
			      (t (selectq (compare-env new-env (car penvs))
				   ;; If you find yourself.  By minimality don't look
				   ;; further.  But get rid of the duplicate.  
				   ;; Note that new should have been set by now.
				   (EQUAL (rplaca penv nil) (return))
				   ;; If you find an earlier element which you 
				   ;; subsume, delete it.  
				   ;; By minimality you won't find yourself,
				   ;; or something that subsumes you.  
				   (SUBSET12 (rplaca penvs nil) (setq flag T))
				   ;; If you find something that subsumes you, forget about
				   ;; yourself.
				   (SUBSET21 (rplaca penv nil) (return)))))))))
	   ;; new is T, unless partial-envs is totally nil.
	   (unless new (setq partial-envs nil) (return)))
	  ;;***** does the following have the same bug as yourself?
	  ;; ******* 99% of the calls to union happen right here ******
	  (T (dolist (env (node-envs node))
	       (dolist (penv partial-envs)
		 (unless (or (null penv)
			     (env-contradictory (setq new-env (2-union-env penv env)))
			     (dolist (known-env known-envs)
			       (if (subset-env? known-env new-env) (return T))))
		   (setq flag nil)	
		   (do ((penvs npartial-envs (cdr penvs)))
		       ((null penvs) (unless flag (push new-env npartial-envs)))
		     (cond ((null (car penvs)))
			   (flag (if (subset-env? new-env (car penvs))
				     (rplaca penvs nil)))
			   (t (selectq (compare-env new-env (car penvs))
				(EQUAL (return))
				(SUBSET12 (rplaca penvs new-env) (setq flag T))
				(SUBSET21 (return)))))))))
	     (setq partial-envs npartial-envs)
	     (unless npartial-envs (return nil)))))
  (fdelqa nil partial-envs))

;;; A specialization of union, when the result is going to be marked inconsistent.
;;; Changes to union-env should perculate to here.
(defun union-for-false (e1 e2 reason &aux result unions1 unions2 count1 count2)
  (cond	((and *union-check*
	      (or (vector-intersection? (env-vector e1) (car (env-cons-env-cache e2)))
		  (vector-intersection? (env-vector e2) (car (env-cons-env-cache e1))))))
	((env-contradictory e1))
	((env-contradictory e2))
	((eq e1 e2) (contradictory-env e1 `(BASE UNION-FOR-FALSE ,reason)))
	((eq e1 *empty-env*) (contradictory-env e2 `(BASE UNION-FOR-FALSE reason)))
	((eq e2 *empty-env*) (contradictory-env e1 `(BASE UNION-FOR-FALSE reason)))
	((= (setq count1 (env-count e1)) 1)
	 (cons-for-false e2 (car (env-assumptions e1)) reason))
	((= (setq count2 (env-count e2)) 1)
	 (cons-for-false e1 (car (env-assumptions e2)) reason))
	;; We could optimize cons-env now****
	;; If complete only one of these tests should be sufficient.

	((if (< count1 count2)
	     (if (vector-subset (env-vector e1) (env-vector e2))
		 (contradictory-env e2 `(BASE UNION-FOR-FALSE reason)))
	     (if (< count2 count1)
		 (if (vector-subset (env-vector e2) (env-vector e1))
		     (contradictory-env e1 `(BASE UNION-FOR-FALSE reason))))))
	;; *** if union-check is enabled this maybe should cache.  But it never returns
	;; T why?
	((not (compatible-env? e1 e2)))
	(*cache-union-flag*
	 (setq unions1 (env-unions e1) unions2 (env-unions e2))
	 (cond ((and unions1
		     unions2
		     (if (> (car unions1) (car unions2))
			 (cdr (assq e1 (cdr unions2)))
			 (cdr (assq e2 (cdr unions1))))))
	       (t (setq result (union-for-false1 e1 e2 reason))
		  (cond (unions1
			 (incf (car unions1))
			 (rplacd unions1 (cons (cons e2 result) (cdr unions1))))
			(t (setf (env-unions e1) (list 1 (cons e2 result)))))
		  (cond (unions2
			 (incf (car unions2))
			 (rplacd unions2 (cons (cons e1 result) (cdr unions2))))
			(t (setf (env-unions e2) (list 1 (cons e1 result)))))
		  result)))
	((union-for-false1 e1 e2 reason))))

;;; Designed really for external calls.
(defun user-general-weave (known-envs new-partial-envs antecedents
			   &optional (node-to-ignore 'IGNORE))
  (dolist (n-a antecedents)
    (cond ((eq n-a node-to-ignore))
	  ((i-out? n-a) (return-from USER-GENERAL-WEAVE known-envs))))
  (general-weave known-envs new-partial-envs antecedents))

;;; General-weave assumes it is not called with known-envs = *empty-env-list* because
;;; that should have been caught earlier.
;;; General weave assumes every antecedent is in, if not this should be trapped earlier.
(defun general-weave (known-envs new-partial-envs antecedents
		      &optional (node-to-ignore 'IGNORE) &aux new-envs flag)

  (cond ;; If any antecedent is known, punt.
	((and known-envs
	      (dolist (n-a antecedents)
		(cond ((eq n-a node-to-ignore))
		      ((known-in-envs? n-a known-envs) (return T)))))
	 known-envs)
			   
	;;; If we smash new-partial-envs what about cdr-coding.
	(t 
	   ;; Efficiency hack which may or may not help.  It saves an initial function
	   ;; call in many cases, and speeds up cases where antecedents are true.
	   (do nil (nil)
	     (if (eq node-to-ignore (car antecedents)) (pop antecedents))
	     (cond ((and (eq (car new-partial-envs) *empty-env*) antecedents)
		    (setq new-partial-envs (node-envs (car antecedents)))
		    (pop antecedents))
		   (t (return nil))))
	   
	   (setq new-partial-envs (fcopylist new-partial-envs))
	   ;; Remove the already known partial-envs.
	   (do ((env new-partial-envs (cdr env)))
	       ((null env))
	     (if (dolist (known known-envs) (if (subset-env? known (car env)) (return T)))
		 (rplaca env nil)
		 (setq flag T)))
	   (cond ((null flag) known-envs)
		 ((setq new-envs
			(nweave antecedents new-partial-envs node-to-ignore known-envs))
		  (setq known-envs (fcopylist known-envs))
		  (dolist (new-env new-envs)
		    (do ((known-env known-envs (cdr known-env)))
			((null known-env))
		      (and (car known-env)
			   (proper-subset-env? new-env (car known-env))
			   (rplaca known-env nil))))
		  (setq known-envs (nconc (fdelqa nil known-envs) new-envs))
		  (values known-envs new-envs))
		 (t known-envs)))))

;;; A version of general-weave specialized for weaving for false.  If this still
;;; takes a lot of time this can be optimized a lot more.
(defun weave-for-false (envs antecedents node-to-ignore)
  (cond ;; If any antecedents are out, punt.
	((dolist (n-a antecedents)
	   (cond ((eq n-a node-to-ignore))
		 ((i-out? n-a) (return T)))))
	(node-to-ignore
	 (dolist (e envs)
	   (weaving-for-false antecedents e node-to-ignore)))
	(t (weaving-for-false antecedents *empty-env* nil))))

;;; It is a theorem that nodes has at least two elements initially.  This goes way out
;;; of its way to avoid calling union-env.  **** The changes here could perculate
;;; to the -c012 version.
;;;**** Changed added 6/23/88 which -c012 doesn't have either: a future node might
;;; become out, that wasn't checked for.  It happens in (dn-queens 4 t) with
;;; *resolve-by-ordered-labeling*.  Some day when optimizing, remove the out? check
;;; and see what actually is causing the problem.
;;; **** unfortunately the above doesn't happen anymore because I changed ADDB so there
;;; may be a bug lurking here.
(defun weaving-for-false (nodes env node-to-ignore &aux next-nodes flag envs)
  ;; Punt if we can't do anything interesting.
  (if (env-contradictory env) (return-from WEAVING-FOR-FALSE nil))
  (prog ()
	start
	;; Skip to the first node which can tell us anything.
	(do nil ((null nodes)
		 ;; This can happen. (justify-node F x T T T) for example.
		 (contradictory-env env `(BASE WEAVING-FOR-FALSE))
		 ;,*justification*))
		 (return-from WEAVING-FOR-FALSE nil))
	  (cond ((i-true? (car nodes)))
		((eq (car nodes) node-to-ignore) (setq node-to-ignore nil))
		;**?? Optimize this case somehow.
		((i-out? (car nodes)) (return-from WEAVING-FOR-FALSE nil))
		(t (return nil)))
	  (setq nodes (cdr nodes)))
	(cond ((neq env *empty-env*))
	      ((null (cdr (node-envs (car nodes))))
	       (setq env (car (node-envs (car nodes))) nodes (cdr nodes))
	       (go start))
	      (t ;;****** also on the last e, why don't we 
		 (dolist (e (node-envs (car nodes)))
		   ;;*** sound, but not optimal.  I added the following kludge, but
		   ;;needs to be thought about.  Really we should check whether e
                   ;; is still part of the label.  We could do a complex check to make
		   ;; sure e is still in the label.****** If a deletion happens
		   ;; **** is the envs updated?????*************
		   ;; **** occurs only with nogood resolution.
		   ;;*** in general after every call to weaving for false we need to
		   ;; to check the envs and nodes we have in or hands.
;		   (if (memq e (node-envs (car nodes)))
		       (weaving-for-false (cdr nodes) e node-to-ignore))
		 (return-from WEAVING-FOR-FALSE nil)))
	;; Now we have a non-empty env, and nodes starting with something interesting.
	(setq next-nodes (cdr nodes))
	loop
	(do nil ((null next-nodes)
		 (unions-for-false env (node-envs (car nodes)) 'IGNORED)
		 ;*justification*)
		 (return-from WEAVING-FOR-FALSE nil))
	  (cond ((i-true? (car next-nodes)))
		((eq (car next-nodes) node-to-ignore) (setq node-to-ignore nil))
		;;**** check this earlier.
		((i-out? (car next-nodes)) (return-from WEAVING-FOR-FALSE nil))
		(t (return nil)))
	  (setq next-nodes (cdr next-nodes)))
	;; Now we are in a situation where we have a non-empty env, and two interesing nodes.
	;; We will have no choice but to do a real union-env.  Lets hope next-nodes is a
	;; singleton.
	(when (null (cdr (node-envs (car next-nodes))))
	  (setq env (union-env env (car (node-envs (car next-nodes)))))
	  (if (env-contradictory env) (return-from WEAVING-FOR-FALSE nil))
	  (setq next-nodes (cdr next-nodes))
	  (go loop))
	;; env is not empty, neither nodes or next-envs is a singleton.  
	(setq flag nil envs nil)
	(dolist (e (node-envs (car nodes)))
	  (unless (or (env-contradictory (setq e (union-env env e)))
		      (and flag (eq e env))
		      (memq e envs))
	    (if (eq e env) (setq flag T) (push e envs))
	    (weaving-for-false next-nodes e node-to-ignore)
	    (if (env-contradictory env) (return-from WEAVING-FOR-FALSE nil))))))

;;; Try to be clever if everything is true.
;;; **** known-envs hack.

(defun nweave (nodes-to-process partial-envs node-to-ignore known-envs &aux npartial-envs)
  ;; Try to handle the extremely simple common case where everything is a singleton,
  ;; and dispatch back to the general case if you lose.
  ;; *** we could have many difference versions of nweave-. for each case.
  ;; *** no need to return to the general general case.
  ;; *** if the general case reduces envs to one env, it could dispatch
  ;; *** to the easy case.
  (cond ((and (null known-envs) (null (cdr partial-envs)))
	 (nweave-simple nodes-to-process (car partial-envs) node-to-ignore))
	((eq (car partial-envs) *empty-env*)
	 (do ((nodes-to-process nodes-to-process (cdr nodes-to-process)))
	     ((null nodes-to-process) partial-envs)
	   (cond ((i-true? (car nodes-to-process)))
		 ((eq (car nodes-to-process) node-to-ignore))
		 (t (setq npartial-envs (fcopylist (n-a-envs (car nodes-to-process))))
		    ;;*** remove subsets function.
		    (when known-envs
		      (do ((tenv npartial-envs (cdr tenv)))
			  ((null tenv))
			(if (dolist (known known-envs)
			      (if (subset-env? known (car tenv)) (return T)))
			    (rplaca tenv nil)))
		      (setq npartial-envs (fdelqa nil npartial-envs)))
		    (if npartial-envs
			(return (weave1 (cdr nodes-to-process) npartial-envs
					node-to-ignore known-envs))
			(return nil))))))
	(t (weave1 nodes-to-process partial-envs node-to-ignore known-envs))))

(defun weave1 (nodes-to-process partial-envs node-to-ignore known-envs
	       &aux new-env npartial-envs flag)
;  (push (cons (length partial-envs) (mapcar #'(lambda (n)
;						(if (neq node-to-ignore n) (length (node-envs n))))
;					    nodes-to-process))
;	foo)

  (dolist (node nodes-to-process)
    (setq npartial-envs nil)
    (cond ((eq node node-to-ignore))
	  ((i-true? node))
	  ;; If the label has exactly one environment, cross product is simpler.
	  ((null (cdr (node-envs node)))
	   (setq partial-envs
		 (weave-single-env (car (node-envs node)) known-envs partial-envs))
	   (unless partial-envs (return nil)))
	  ((null (cdr partial-envs))
	   (setq partial-envs
		 (safe-weave-single-env (car partial-envs) known-envs (node-envs node)))
	   (unless partial-envs (return nil)))
	  ;;***** does the following have the same bug as yourself?
	  ;; ******* 99% of the calls to union happen right here ******
	  (T (dolist (env (node-envs node))
	       (dolist (penv partial-envs)
		 (unless (or (null penv)
			     (env-contradictory (setq new-env
						      (union-env3 known-envs penv env))))
		   (setq flag nil)	
		   (do ((penvs npartial-envs (cdr penvs)))
		       ((null penvs) (unless flag (push new-env npartial-envs)))
		     (cond ((null (car penvs)))
			   (flag (if (subset-env? new-env (car penvs))
				     (rplaca penvs nil)))
			   (t (selectq (compare-env new-env (car penvs))
				(EQUAL (return))
				(SUBSET12 (rplaca penvs new-env) (setq flag T))
				(SUBSET21 (return)))))))))
	     (setq partial-envs npartial-envs)
	     (unless npartial-envs (return nil)))))
  (fdelqa nil partial-envs))

;;; If much time is spent here, make a version of weave-single-env which wins here.
(defun safe-weave-single-env (env known-envs partial-envs &aux safe-partial-envs)
  (dolist (penv partial-envs)
    (unless (dolist (known known-envs) (if (subset-env? known penv) (return T)))
      (push penv safe-partial-envs)))
  (weave-single-env env known-envs safe-partial-envs))

;; This loop updates the partial-envs list in place.  This works because the
;; new partial-envs can't be any bigger than the original one.
(defun old-weave-single-env (env known-envs partial-envs &aux new-env flag new)
  (do ((penv partial-envs (cdr penv)))
      ((null penv))
    (cond ((null (car penv)))
	  ;; If the new union is inconsistent, delete it from the partial result.
	  ((env-contradictory
	     (setq new-env (union-env3 known-envs (car penv) env)))
	   (rplaca penv nil))
	  ;; Otherwise, scan the updated prefix of partial-envs.  When you
	  ;; find old self, replace it with your new self.
	  (t (setq flag nil)
	     (do ((penvs partial-envs (cdr penvs)))
		 ((eq penvs penv) (setq new T) (rplaca penvs new-env))
	       (cond ((null (car penvs)))
		     ;; This is based on the theorem that if the prefix is
		     ;; always minimal, then if you have found an element prefix
		     ;; which you subsume, you cannot find yourself, nor any
		     ;; element which subsumes you.
		     (flag (if (subset-env? new-env (car penvs))
			       (rplaca penvs nil)))
		     ;; Scan the prefix, to ensure that minimality is
		     ;; maintained.  If you ever find an element you subsume,
		     ;; then the much easier previous check is faster.
		     (t (selectq (compare-env new-env (car penvs))
			  ;; If you find yourself.  By minimality don't look
			  ;; further.  But get rid of the duplicate.  
			  ;; Note that new should have been set by now.
			  (EQUAL (rplaca penv nil) (return))
			  ;; If you find an earlier element which you 
			  ;; subsume, delete it.  
			  ;; By minimality you won't find yourself,
			  ;; or something that subsumes you.  
			  (SUBSET12 (rplaca penvs nil) (setq flag T))
			  ;; If you find something that subsumes you, forget about
			  ;; yourself.
			  (SUBSET21 (rplaca penv nil) (return)))))))))
  ;; new is T, unless partial-envs is totally nil.
  (unless new (setq partial-envs nil))
  partial-envs)

;;; ***** idea if union-env3 result here is a subset only check for eq?

(defun weave-single-env (env known-envs partial-envs)
  (dolist (penv partial-envs)
    (when (and penv (subset-env? penv env))
      (dolist (known known-envs)
	(if (subset-env? known env)
	    (return-from WEAVE-SINGLE-ENV nil)))
      (return-from WEAVE-SINGLE-ENV (list env))))
  (cond ((= (env-count env) 1)
	 (new-weave-assumption (car (env-assumptions env)) known-envs partial-envs))
	(t (old-weave-single-env env known-envs partial-envs))))

;;; Prefix always contained assumption.
(defun new-weave-assumption (assumption known-envs partial-envs &aux non new last-non new-env)
  (do ((penv partial-envs (cdr penv))
       (ppenv nil (if recycle ppenv penv))
       (recycle nil nil))
      ((null penv)
       (cond ((null new) nil)
	     (non (rplacd last-non partial-envs) non)
	     (t partial-envs)))
    (cond ((null (car penv)))
	  ((vector-member assumption (env-vector (car penv)))
	   (setq new T)
	   (do ((non non (cdr non))
		(pnon nil non))
	       ((null non))
	     (and (car non)
		  (subset-env? (car non) (car penv))
		  (rplaca non nil))
	     (if (eq last-non non) (return nil))))
	  ((or (env-contradictory (setq new-env (cons-env (car penv) assumption)))
	       (dolist (known known-envs) (if (subset-env? known new-env) (return T))))
	   (rplaca penv nil))
	  ;; Now we have another non, its prime with respect to other nons, but
	  ;; it might be a superset of the prefix.
	  (t (do ((penvs partial-envs (cdr penvs)))
		 ((eq penvs penv)
		  (setq new T)
		  (cond (last-non (rplacd last-non penv)
				  (setq last-non penv))
			(t (setq non penv last-non penv)))
		  (setq recycle T)
		  (if ppenv (rplacd ppenv (cdr penv)) (setq partial-envs (cdr penv)))
		  (rplaca last-non new-env)
		  )
	       (when (and (car penvs) (subset-env? (car penvs) new-env))
		 (rplaca penv nil)
		 (return nil)))))))

;;; creates a set of environments by conjoining all the environments of the nodes-to-process
;;; along with the new-partial-environment and adds these to *new-envs*
;;; This tries to be efficient.***** be cleverer about partial-envs.

;;; This handles the simple case of weaving where there are no known-envs, and
;;; all the antecedents have singleton environments.  This happens very commonly
;;; in justifications for the *contra-node*.
(defun nweave-simple (nodes-to-process env node-to-ignore
				       &aux node nenv env1 env2 envs comp last-nil oenvs)
  (do ((nodes-to-process nodes-to-process (cdr nodes-to-process)))
      ;; We could with crockery save this cons.
      ((null nodes-to-process) (list env))
    ;; If the node is to be ignored, ignore it.
    (unless (eq node-to-ignore (setq node (car nodes-to-process)))
      (setq nenv (car (n-a-envs node)))
      (cond ;; If the node is true, it contributes nothing.
	    ((eq nenv *empty-env*))
	    ;; If the next node has a simple label, its easy.
 	    ((null (cdr (node-envs node)))
	     (setq env (union-env env nenv))
	     (if (env-contradictory env) (return nil)))
	    ;; Here things get slightly dicey.  Lets hope that only one of the
	    ;; envs we have to construct survive.
	    (t (dolist (nenv (node-envs node))
		 (setq env1 (union-env env nenv))
		 (cond ;; If the result is contradictory we are in luck.
		       ((env-contradictory env1))
		       ;; If we haven't got any consistent envs yet, we are in luck.
		       ((null env2) (setq env2 env1))
		       ;; Things are not hopless yet, they might compare.
		       ((and (null envs) (setq comp (compare-env env1 env2)))
			(selectq comp
			  (EQUAL) ;;Damn Cl.
			  (SUBSET21)
			  (SUBSET12 (setq env2 env1))))
		       ;; The screw case, things are no longer simple.  We just have
		       ;; to byte the bullet and finish our work, and then trap
		       ;; out to the general case again.
		       ((null envs) (setq envs (list env1 env2)))
		       ;; We have to check for subsumption, in the usual way.
		       ;; Same old theorems apply.
		       (t (setq oenvs envs last-nil nil)
			  (do ((envs envs (cdr envs)))
			      ((null envs)
			       (if last-nil
				   (rplaca last-nil env1)
				   (push env1 oenvs)))
			    (cond ((null (car envs)) (setq last-nil envs))
				  (t (selectq (compare-env env1 (car envs))
				       ;; If the label is minimal, you could not have hacked
				       ;; envs at all yet.  This env does nothing.
				       (EQUAL (return))
				       (SUBSET12 (rplaca envs env1)
						 (do ((envs (cdr envs) (cdr envs)))
						     ((null envs))
						   ;; Can't find yourself if label was min.
						   (when (and (car envs)
						       (proper-subset-env? env1 (car envs)))
						     (rplaca envs nil)
						     (setq last-nil envs)))
						 (when last-nil
						   (fdelqa nil oenvs)
						   (unless (cdr oenvs)
						     (setq env2 (car oenvs) oenvs nil)
						   (return))))
				       (SUBSET21 (return))))))
			  (setq envs oenvs))))
	       ;; If envs is non-nil at this point, we have lost and can't use
	       ;; the simple approach any more: There is more than one env in the
	       ;; label so far.  Thus, we dispatch back to the general case.
	       (cond ((null envs) (if env2 (setq env env2 env2 nil) (return nil)))
		     ;; We don't have to do the general case if no nodes left.
		     ((dolist (node (cdr nodes-to-process) T)
			(cond ((eq node node-to-ignore))
			      ((true? node))
			      (t (return nil))))
		      (return envs))
		     (t (return
			  (nweave (cdr nodes-to-process) envs node-to-ignore nil)))))))))

;;; Replace the nodes envs with a new set.  If the status changes, and we care, i.e., there
;;; are pending rules on the node, start building a pending queue.  Due to the
;;; large number of bugs, this makes sure that the changed envs are really useful.

;;; Returns T if santization is required....not anymore.

;;; The case of new-envs = empty-env could be optimized.****
(defun change-node-envs (node new-envs &aux old-envs out-informative in-informative)
  ;;**** why doesn't this first option return T if contradictory environments are discoverd?
  (cond ((node-contradictory node)
	 ;;*** we have to have a carefull flag or something to leave this behind.
	 (contradictory-envs new-envs
			     '(BASE LABEL-FOR-FALSE2)))
;			     (list 'BASE
;				   'LABEL-FOR-FALSE2
;				   node new-envs
;				   justification
;				   (mapcar #'(lambda (n-a) (fcopylist (n-a-envs n-a)))
;					   (cddr justification))))
	
	(t (setq old-envs (node-envs node))
	   (setf (node-envs node) new-envs)
	   ;; If an env goes away, unwire the node.
	   (dolist (old-env old-envs)
	     (unless (or (memq old-env new-envs) (env-contradictory old-env))
	       (setq out-informative T)
	       (setf (env-nodes old-env) (fdelq1 node (env-nodes old-env)))))
	   ;; If a new-env comes in, wire the node.
	   (dolist (new-env new-envs)
	     (unless (memq new-env old-envs)
	       (setq in-informative T)
	       (incf *update-counter*)
	       ;; Don't bother keeping two-way pointers for envs.
	       (unless (eq new-env *empty-env*)
		 (push node (env-nodes new-env)))))
	   ;; **** make more efficient by just saving the cons.****
	   (unless (or in-informative out-informative (eq (car new-envs) *empty-env*))
	     (error "Uninformative node update"))
	   (when in-informative
	     ;; We may have to queue these at a higher level sometime, because some of the
	     ;; labeling algorithm may depend on labels not changing below it, and
	     ;; hence have a queue just like *going-nodes*.
	     (if (eq (car new-envs) *empty-env*) (new-true-node node))
	     (dolist (c (node-env-consumers node)) (add-node-consumer-to-node c node))
	     (setf (node-env-consumers node) nil))
	   (cond (old-envs (cond ((null new-envs) (out-node node))
				 ((and (node-enqueued? node) in-informative)
				  (enqueue-node node))
				 ))
		 (new-envs 
		  (setf (node-status node)
			(if (has-consumers? node)
			    (selectq (node-status node)
			      (GOING-OUT 'STAYING-IN)
			      (STAYING-OUT 'GOING-IN)
			      (OUT (push node *going-nodes*) 'GOING-IN)
			      ((IN GOING-IN STAYING-IN)
			       (error "Consumer node was already ~A?" (node-status node)))
			      (T (error "Impossible node status ~A"	
					(node-status node))))
			    'IN))))))

  nil)

;;; Tell the TMS that the given node is contradictory.
(defun contradictory-node (node reason &aux *going-nodes*)
  reason
  (unless (node-contradictory node)
    (propagate-false node)
    (if *going-nodes* (process-changed-nodes))))

;;; This propagates TRUE/FALSE through the entire network without doing any label
;;; updating in between at all.  Assuming this runs underneath *going-nodes*.
(defvar *set-nodes* nil)

(defun new-true-node (node &aux *set-nodes*)
  (propagate-true-2 node)
  (set-nodes))

(defun propagate-true (node &aux *set-nodes*)
  (propagate-true-1 node)
  (set-nodes))

(defun propagate-false (node &aux *set-nodes*)
  (propagate-false-1 node)
  (set-nodes))

;;; Not optimized for assumptions being set at all.  This is very wierd and needs more thought,
;;; the envs data structure will be a bit inconsistent until set-nodes finishes.
;;; Also, *set-nodes* should be a set of symbols, and not nodes.
(defun set-nodes ()
  (dolist (n *set-nodes*)
    (when (n-a-contradictory n)
      (do nil ((null (node-envs n)))
	(contradictory-env (car (node-envs n)) '(BASE SET-NODES)))))	
  (dolist (n *set-nodes*)
    (unless (n-a-contradictory n)
      (update-node1 n *empty-env-list* *empty-env-list*))))

(defun propagate-true-1 (node)
  (if (node-contradictory node) (throw 'CONTRADICTION 'FAIL))
  (unless (i-true? node)
    (push node *set-nodes*)
    (setf (n-a-envs node) *empty-env-list*)
    (propagate-true-2 node)))

;;; This is called when a node gets an empty environment.  If the node participates in
;;; a justification for false, false might be back propagatable.  Again, I don't know
;;; whether this is worth it.

(defun propagate-true-2 (node)
    (when (n-a-neg node) (propagate-false-1 (n-a-neg node)))
    (when (assumption? node)
      ;; Remove all assumption disjunctions mentioning it.
      (dolist (or (assumption-ors node))
	(excise-or or)
	;;*** perhaps this should be thought about some more.  
	(setf (disjunction-satisfied or) T)))
    (dolist (class (node-classes node))
      (when (class-exclusive class)
	(dolist (other (class-nodes class))
	  (unless (eq other node)
	    (propagate-false-1 other)))))
    (dolist (c (node-consequents node))
      (if (n-a-contradictory (car c)) (contradictory-just-1 (cddr c))))
    (dolist (c (n-a-nclauses node)) (propagate-clause node c)))

(defun propagate-false-1 (node &aux open)
  (if (i-true? node) (throw 'CONTRADICTION 'FAIL))
  (unless (node-contradictory node)
    (setf (node-contradictory node) T)
    ;; Remove all useless justifications in which this node participates.
    ;; Consequent looks like (node . justification)
    (dolist (cnsqnt (node-consequents node))
      (setf (node-justifications (car cnsqnt))
	    (fdelq1 (cdr cnsqnt) (node-justifications (car cnsqnt))))
      (dolist (antecedent (cddr cnsqnt))
	(unless (eq node antecedent)
	  (setf (node-consequents antecedent) (fdelq1 cnsqnt (node-consequents antecedent))))))
    (setf (node-consequents node) nil)
    (push node *set-nodes*)
    (when (n-a-neg node) (propagate-true-1 (n-a-neg node)))
    (dolist (class (node-classes node))
      (when (class-oneof class)			; Redundant because clause is always there?
	(setq open nil)
	(dolist (node (class-nodes class) (if open (propagate-true-1 open))))
	  (cond ((i-false? node))
		((i-true? node))
		(open (return nil))
		(t (setq open node)))))
    (dolist (j (node-justifications node)) (contradictory-just-1 (cdr j)))
    (dolist (c (n-a-pclauses node)) (propagate-clause node c))))

;;; Given a justification for a contradictory node, this propagates it backwards.  This
;;; is incomplete as well, but it is an efficiency hack.  If the single antecedent is
;;; an assumption it should already be taken care of.
(defun contradictory-just-1 (antecedents &aux only-supporter)
  (unless (dolist (supporter antecedents)
	    (cond ((i-true? supporter))
		  ((n-a-contradictory supporter) (return T))
		  (only-supporter (return T))
		  (t (setq only-supporter supporter))))
    (unless only-supporter (throw 'CONTRADICTION 'FAIL))
    (propagate-false-1 only-supporter)))

;;; This is incorrect if negatives don't exist.  WHICH THEY SOMETIMES DON'T FOR ASSUMPTIONS.
;;; ****
(defun propagate-clause (node clause &aux open)
  node
  ;; All negatives must be true, or nothing can be concluded.
  (dolist (n (first clause))
    (cond ((i-true? n))
	  (open (return-from PROPAGATE-CLAUSE nil))
	  (t (setq open n))))
  ;; We have an open negatively occuring literal.
  (dolist (p (second clause))
    (cond ((i-false? p))
	  (open (return-from PROPAGATE-CLAUSE nil))
	  (t (setq open p))))
  ;; We have an open literal, all other negatives are true, and all other positives are false,
  ;; so open MUST be true.
  (if (memq open (first clause))
      ;; Open literal occurs negatively, it must be false.
      (propagate-false-1 open)
      ;; Open literal occurs positively, it must be true.
      (propagate-true-1 open)))

;;; True-assumption is called when the ATMS discovers some disjunction of assumptons
;;; is reduced to a singleton.

;;; Peter Struss claims this is probably
;;; extremely inefficient.****************
;;; ******* rehasing should flush these from the table.
;;; ******** Also the justify-node will now detect the smaller nogoods.  In actual
;;; fact it may be possib
(defun true-assumption (assumption informant)
  ;; Make the assumption universally true.
  (justify-node assumption (list informant))
  ;; Mark its negation as false.
  ;; Soon to be obsolete I think.****
  (when (or *resolve-by-labeling* *resolve-by-ordered-labeling*)
    (if (assumption-negation assumption)
	(contradictory-node (assumption-negation assumption) '(TRUE-ASSUMPTION))))
  )

