;;; -*- Mode: LISP; Syntax: Common-lisp; Package: TMS; Base: 10; Patch-File: Yes; -*-

(in-package 'TMS) 

   ;;; Basic problem solving operation is to justify one node in terms of others.
   ;;; It does not return a useful value.

   ;;; This algorithm could stand a lot more thought if the functionality is right.
   ;;; Loops are ok because they have to die out.
   ;;; Note that the same node may be reexamined many times.
   ;;; Maybe there should be a breadth-first queue.

   ;;; A justification looks like (user-supplied-informant . antecedent-nodes).
   (defun justify-node (node justification &optional dont-remove-duplicates
			&aux *going-nodes* consequent j-consumers out clean n ptr)
     (incf *justification-count*)
     (when *trace-file*
	 (format *trace-file* "~%J-N ")
	 (trace-node node)
	 (format *trace-file* "(")
	 (mapc #'(lambda (n-a) (trace-node n-a)) (cdr justification))
	 (format *trace-file* ")"))
     ;; If the consequent is true, chuck the justification.
     (if (i-true? node) (return-from JUSTIFY-NODE nil))

     (do ((antecedents (cdr justification) (cdr antecedents)))
	 ((null antecedents)
	  (if clean (rplacd justification (fdelqa nil (cdr justification)))))
       (when (setq n (car antecedents))
	 (if (i-false? n) (return-from JUSTIFY-NODE nil))
	 (or out (n-a-envs n) (setq out T))
	 ;; Rare, so this optimizes the case where there are no duplicates.
	 (or out dont-remove-duplicates
	     (when (setq ptr (memq n (cdr antecedents)))
	       (do nil (nil)
		 (rplaca ptr nil)
		 (unless (setq ptr (memq n (cdr ptr))) (setq clean T) (return nil)))))))
     (push justification (node-justifications node))
     (setq consequent (cons node justification))
     (dolist (antecedent (cdr justification)) (push consequent (n-a-consequents antecedent)))
     (cond ((node-contradictory node)
	    (contradictory-just (cdr justification))
	    (weave-for-false nil (cdr justification) nil))
	   (out (return-from JUSTIFY-NODE nil))
	   (t (multiple-value-bind (known-envs new-envs)
		  (general-weave (node-envs node) *empty-env-list* (cdr justification))
		(when new-envs (update-node1 node new-envs known-envs)))))
     (if *going-nodes* (process-changed-nodes))
     (when (setq j-consumers (node-justification-consumers node))
       (dolist (c j-consumers) (add-node-consumer-to-node c node))
       (setf (node-justification-consumers node) nil))
     )


   (defun contradictory-just (antecedents &aux *set-nodes*)
     (contradictory-just-1 antecedents)
     (set-nodes))
