;;; -*- Syntax:Common-Lisp; Mode: LISP; Package: (TMS Lisp 1000.); Base: 10. -*-

"(c) Copyright 1986, 1987, 1988 Xerox Corporation.  All rights reserved.  Subject to
the following conditions, permission is granted to use and copy this
software and to prepare derivative works:  Such use, copying or
preparation of derivative works must be for non-commercial research or
educational purposes; each copy or derivative work must include this
copyright notice in full; a copy of each completed derivative work must
be returned to:  DEKLEER@XEROX.COM (Arpanet) or Johan de Kleer,
Xerox PARC, 3333 Coyote Hill Road, Palo Alto, CA 94304.  This software
is made available AS IS, and Xerox Corporation makes no warranty about
the software or its performance."

(in-package 'tms)


(defun walk-tree (function tree) (if tree (walk-tree-1 function tree)))

(defun walk-tree-1 (function tree)
  (cond ((atom tree) (funcall function tree))
	((dolist (entry tree) (walk-tree-1 function (cdr entry))))))
  
(defun count-tree (tree &aux (result 0))
  (walk-tree #'(lambda (ignore) (incf result)) tree)
  result)

;;; Allowable-misses is the number of assumptions we can pass not contained in
;;; the assumption set.
;;; Returns T if current tree is completely deleted.
(defun subsumed-tree (function allowable-misses as tree &aux au)
  (cond ((null as) (walk-tree-1 function tree) T)
	((= 0 allowable-misses) (subsumed-tree-remove as tree function))
	(t (setq au (assumption-unique (car as)))
	   (do ((subtree tree)
		(delete nil)
		(entry nil)
		(previous nil))
	       ((null subtree))
	     (setq entry (car subtree))
	     ;; If the tree contains exactly the next assumption in order, we're golden.
	     (cond ((= au (car entry))
		    (setq delete
			  (subsumed-tree function allowable-misses (cdr as) (cdr entry))))
		   ;; We MUST find the assumption, or else things are pointless.
		   ;; Unique-id's goes down within the tree, so abort if we've gone too far.
		   ;; Wish tree were sorted for this ...
		   ((< (car entry) au))
		   ;; Take the current tree assumption as a miss, keep on looking.
		   (t (setq delete
			    (subsumed-tree function (1- allowable-misses)
					   as (cdr entry)))))
	     (cond (delete (cond ((null (cdr tree)) (return T))
				 (previous (rplacd previous (cdr subtree))
					   (setq subtree (cdr subtree)))
				 (t (rplaca subtree (cadr subtree))
				    (rplacd subtree (cddr subtree))))
			   (setq delete nil))
		   (t (setq previous subtree subtree (cdr subtree))))))))

(defun subsumed-tree-remove (as tree function &aux au)
  (setq au (assumption-unique (car as)))
  (do ((subtree tree (cdr subtree))
       (previous nil subtree))
      ((null subtree) nil)
    (cond ((not (= au (caar subtree))))
	  ((or (when (not (listp (cdar subtree)))
		 (funcall function (cdar subtree))
		 T)
	       (subsumed-tree-remove (cdr as) (cdar subtree) function))
	   ;; Delete this subtree, except if its the only one.
	   (cond ((null (cdr tree)) (return T))
		 (previous (rplacd previous (cdr subtree)))
		 (t (rplaca subtree (cadr subtree))
		    (rplacd subtree (cddr subtree))))
	   (return nil))
	  (t (return nil)))))

;;; Note that there is an overriding assumption here that the set of
;;; assumptions are stored in order: descending unique-id's.

;;; ASSQ= or something whihc uses unique-ids'  Put unique-ids only in tree is
;;; probably faster anywyay???

;;; This can be cleverer by going down both simultaneously.
;;; Tree always looks like either a singleton env or a list of (<assumption> . tree)
;;; This has the built-in assumption everything is of the same length.

;;; This is somewhat bummed.
(defun subsumed-by-tree-1 (allowable-misses as tree &aux slot)
  (if (= 0 allowable-misses)
      ;; This changes into a straight lookup, if list runs out we succeeded.
      (dolist (a as)
	(or a (throw 'SUBSUMED tree))
	(unless (setq slot (assq a tree)) (return nil))
	(setq tree (cdr slot)))
      (subsumed-by-tree-3 allowable-misses as tree)))

(defun subsumed-by-tree-3 (allowable-misses as tree &aux slot)
  (dolist (entry tree)
    (do* ((as as (cdr as))
	  (allowable-misses allowable-misses (1- allowable-misses)))
	 ((= allowable-misses 0)
	  (when (= (car as) (car entry))
	    (setq tree (cdr entry))
	    (dolist (a (cdr as))
	      (or a (throw 'SUBSUMED tree))
	      (unless (setq slot (assq a tree)) (return nil))
	      (setq tree (cdr slot)))))
      (when (= (car entry) (car as))
	(unless (listp (cdr entry)) (throw 'SUBSUMED (cdr entry)))
	(subsumed-by-tree-1 allowable-misses (cdr as) (cdr entry))
	(return nil))
      (when (> (car entry) (car as))
	(return nil)))))

;;; Exactly like subsumed-by-tree-1 except that we know (- as a) itself is known to be
;;; not subsumed by the tree.  Therefore anything we look at must have a in it.
(defun subsumed-by-tree-2 (allowable-misses a as tree &aux slot it)
  (cond ((= 0 allowable-misses)
	 ;; This changes into a straight lookup.
	 (do nil (nil)
	   (unless (setq slot (assq (assumption-unique (car as)) tree)) (return nil))
	   (unless (listp (setq tree (cdr slot))) (return tree))
	   (unless (setq as (cdr as)) (error "Bad tree"))))
	(t (dolist (entry tree)
	     (do ((as as (cdr as))
		  (au)
		  (allowable-misses allowable-misses (1- allowable-misses)))
		 (nil)
	       (setq au (assumption-unique (car as)))
	       (when (= au (car entry))
		 (if (listp (cdr entry))
		     (if (setq it
			       (subsumed-by-tree-2 allowable-misses
						   (and a (not (= au a)) a)
						   (cdr as)
						   (cdr entry)))
			 (return-from SUBSUMED-BY-TREE-2 it)
			 (return nil))
		     (return-from SUBSUMED-BY-TREE-2 (cdr entry))))
	       (when (> (car entry) au)
		 (return nil))
	       ;; **** Is this right---maybe a should be in it.
	       (if (eql au a) (return nil))
	       (if (= 0 allowable-misses) (return nil))
	       )))))

;;; Convert a list into a descrimination net.  (This assumes assumptions are ordered!?).
(defun tree-envs (l &aux tree)
  (when l
    (setq tree (insert-tree (car l) tree))
    (dolist (e (cdr l)) (insert-tree e tree)))
  tree)

;;; ASSQ is fast, but if there are lots of assumptios in the list we can do
;;; order it by unique-id.
;;; Remember all entries are of the same length!!

(defun insert-tree (e tree &aux as slot)
  (setq as (env-assumptions e))
  (prog  nil
	 (unless tree (return (build-tree as e)))
    loop (cond ((null (setq slot (assq (assumption-unique (car as)) tree)))
		(setq slot (build-tree as e))
		(rplacd slot (cdr tree))
		(rplacd tree slot)
		(return tree))
	       ((listp (cdr slot))
		(setq tree (cdr slot) as (cdr as))
		(go loop))
	       (t (error "Inserting same env twice")))))

(defun build-tree (as e)
  (if (null as) e (ncons (cons (assumption-unique (car as)) (build-tree (cdr as) e)))))

;;; Return T only when tree becomes empty.
;;; Assumes everything is the same length.
(defun remove-tree (e tree)
  (remove-tree-1 (env-assumptions e) tree))	

;;; ASSQ might be faster?***
;;; Returns T if as was the only thing found in the current tree.
(defun remove-tree-1 (as tree &aux au)
  (setq au (assumption-unique (car as)))
  (do ((subtree tree (cdr subtree))
       (previous nil subtree))
      ((null subtree) (error "Env not in tree"))
    (cond ((not (= au (caar subtree))))
	  ((or (not (listp (cdar subtree))) (remove-tree-1 (cdr as) (cdar subtree)))
	   ;; Delete this subtree, except if its the only one.
	   (cond ((null (cdr tree)) (return T))
		 (previous (rplacd previous (cdr subtree)))
		 (t (rplaca subtree (cadr subtree))
		    (rplacd subtree (cddr subtree))))
	   (return nil))
	  (t (return nil)))))



;;; Return all the environments in the tree.
(defun collect-tree (tree &aux result)
  (walk-tree #'(lambda (env) (push env result)) tree)
  result)

(defvar *result*)

;;; Pure diagnostics.
(defun testr (allowable-misses as tree &aux *result*)
  (find-supersets-in-tree allowable-misses as tree)
  *result*)

(defun qwalk (tree)
  (if (listp tree) (dolist (entry tree) (qwalk (cdr entry))) (push tree *result*)))

(defun find-supersets-in-tree (allowable-misses as tree &aux slot)
  (cond ((null as) (qwalk tree))
	((= 0 allowable-misses)
	 ;; This changes into a straight lookup.
	 (do nil (nil)
	   (unless (setq slot (assq (assumption-unique (car as)) tree)) (return nil))
	   (unless (listp (setq tree (cdr slot))) (push tree *result*) (return tree))
	   (unless (setq as (cdr as)) (error "Bad tree"))))
	(t (dolist (entry tree)
	           ;; If the tree contains exactly the next assumption in order, we're golden.
	     (cond ((= (assumption-unique (car as)) (car entry))
		    (find-supersets-in-tree allowable-misses (cdr as) (cdr entry)))
		   ;; We MUST find the assumption, or else things are pointless.
		   ;; Unique-id's goes down within the tree, so abort if we've gone too far.
		   ;; Wish tree were sorted for this ...
		   ((< (car entry) (assumption-unique (car as))))
		   ;; Take the current tree assumption as a miss, keep on looking.
		   (t (find-supersets-in-tree (1- allowable-misses) as (cdr entry))))))))